"""Numpyro inference engines for prophet models.

The classes in this module take a model, the data and perform inference using Numpyro.
"""

import jax
from skbase.base import BaseObject


class BaseInferenceEngine(BaseObject):
    """
    Class representing an inference engine for a given model.

    Parameters
    ----------
    model : Callable
        The model to be used for inference.
    rng_key : Optional[jax.random.PRNGKey]
        The random number generator key. If not provided, a default key with value 0
        will be used.

    Attributes
    ----------
    model : Callable
        The model used for inference.
    rng_key : jax.random.PRNGKey
        The random number generator key.
    """

    _tags = {
        "object_type": "inference_engine",
    }

    def __init__(self, rng_key=None):
        self.rng_key = rng_key

        if rng_key is None:
            rng_key = jax.random.PRNGKey(0)
        self._rng_key = rng_key

    # pragma: no cover
    def infer(self, model, **kwargs):
        """
        Perform inference using the specified model.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        The result of the inference.
        """
        self.model_ = model
        self._infer(**kwargs)

    # pragma: no cover
    def _infer(self, **kwargs):  # pragma: no cover
        """
        Perform inference using the specified model.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        The result of the inference.
        """
        raise NotImplementedError("infer method must be implemented in subclass")

    # pragma: no cover
    def predict(self, **kwargs):
        """
        Generate predictions using the specified model.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        The predictions generated by the model.
        """
        return self._predict(**kwargs)

    # pragma: no cover
    def _predict(self, **kwargs):  # pragma: no cover
        """
        Generate predictions using the specified model.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        The predictions generated by the model.
        """
        raise NotImplementedError("predict method must be implemented in subclass")



from typing import Any, Dict, Optional, Callable

import jax.numpy as jnp
import pandas as pd

import numpyro
import numpyro.distributions as dist
from prophetverse.effects.base import BaseEffect
from prophetverse.utils.frame_to_array import series_to_tensor_or_array
from prophetverse.distributions import GammaReparametrized, BetaReparametrized
from prophetverse.effects.target.base import BaseTargetEffect


def _do_nothing(x):
    return x


class TargetLikelihood(BaseTargetEffect):
    """Base class for effects."""

    _tags = {
        # Supports multivariate data? Can this
        # Effect be used with Multiariate prophet?
        "hierarchical_prophet_compliant": True,
        # If no columns are found, should
        # _predict be skipped?
        "requires_X": False,
        # Should only the indexes related to the forecasting horizon be passed to
        # _transform?
        "filter_indexes_with_forecating_horizon_at_transform": True,
        # "capability:panel": True,
    }

    def __init__(
        self,
        noise_scale=0.05,
        link_function=None,
        likelihood_func=None,
    ):
        self.link_function = link_function
        self.noise_scale = noise_scale
        self.likelihood_func = likelihood_func
        super().__init__()

    def _fit(self, y, X, scale=1):
        self.likelihood_distribution_ = (
            dist.Normal if self.likelihood_func is None else self.likelihood_func
        )
        self.scale_ = scale

    def _predict(
        self,
        data: Any,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        params : Dict[str, jnp.ndarray]
            A dictionary containing the sampled parameters of the effect.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """

        mean = self._compute_mean(predicted_effects)
        mean = numpyro.deterministic("mean", mean)
        noise_scale = numpyro.sample("noise_scale", dist.HalfNormal(self.noise_scale))

        if mean.ndim <= 2:
            mean = mean.reshape((-1, 1))

            with numpyro.plate("data", len(mean), dim=-2):
                numpyro.sample(
                    "obs",
                    self.likelihood_distribution_(mean, noise_scale),
                    obs=data,
                )
        else:
            with numpyro.plate("series", mean.shape[0], dim=-3):
                with numpyro.plate("data", mean.shape[1], dim=-2):
                    numpyro.sample(
                        "obs",
                        self.likelihood_distribution_(mean, noise_scale),
                        obs=data,
                    )
        return jnp.zeros_like(mean)

    def _compute_mean(self, predicted_effects: Dict[str, jnp.ndarray]) -> jnp.ndarray:
        mean = 0
        for effect_name, effect in predicted_effects.items():
            # Skip latent effects (effects whose names start with "latent/")
            if not effect_name.startswith("latent/"):
                mean += effect

        mean = self.link_function(mean) if self.link_function else mean
        return mean


class NormalTargetLikelihood(TargetLikelihood):
    def __init__(
        self,
        noise_scale=0.05,
    ):
        super().__init__(
            noise_scale, link_function=_do_nothing, likelihood_func=dist.Normal
        )


class GammaTargetLikelihood(TargetLikelihood):
    def __init__(
        self,
        noise_scale=0.05,
        epsilon=1e-5,
    ):
        self.epsilon = epsilon
        link_function = _build_positive_smooth_clipper(epsilon)
        super().__init__(
            noise_scale,
            link_function=link_function,
            likelihood_func=GammaReparametrized,
        )


class NegativeBinomialTargetLikelihood(TargetLikelihood):

    discrete_support = True

    def __init__(
        self,
        noise_scale=0.05,
        epsilon=1e-5,
    ):
        self.epsilon = epsilon
        link_function = _build_positive_smooth_clipper(epsilon)
        super().__init__(
            noise_scale,
            link_function=link_function,
            likelihood_func=dist.NegativeBinomial2,
        )

    def _predict(self, data, predicted_effects, *args, **kwargs):
        y = data

        mean = self._compute_mean(predicted_effects) * self.scale_
        mean = numpyro.deterministic("mean", mean)
        noise_scale = numpyro.sample("noise_scale", dist.HalfNormal(self.noise_scale))

        with numpyro.plate("data", len(mean), dim=-2):
            numpyro.sample(
                "obs",
                dist.NegativeBinomial2(
                    mean.reshape((-1, 1)), noise_scale * self.scale_
                ),
                obs=y,
            )

        return jnp.zeros_like(mean)


def _build_positive_smooth_clipper(
    smooth_threshold: float, threshold: float = 1e-10
) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Force the values of x to be positive.

    Applies a smooth threshold to the values of x to force positive-only outputs.
    Further clips the values of x to avoid zeros due to numerical precision.

    Parameters
    ----------
    smooth_threshold : float
        The threshold value for the exponential function.
    threshold : float, optional
        The threshold value for clipping, by default 1e-10.

    Returns
    -------
    jnp.ndarray
        The transformed array.
    """

    def _to_positive(x):
        return jnp.clip(
            jnp.where(
                x < smooth_threshold,
                jnp.exp(x - smooth_threshold) * smooth_threshold,
                x,
            ),
            threshold,
            None,
        )

    return _to_positive


class BetaTargetLikelihood(TargetLikelihood):
    def __init__(
        self,
        noise_scale=0.05,
        epsilon=1e-5,
    ):
        """Beta likelihood effect.

        Parameters
        ----------
        epsilon : float
            Numerical slack to keep means/factors away from the boundaries.
        """

        self.epsilon = epsilon

        link_function = lambda x: jnp.clip(x, epsilon, 1 - epsilon)
        super().__init__(
            noise_scale=noise_scale,  # not used for beta now
            link_function=link_function,
            likelihood_func=BetaReparametrized,
        )

    def _predict(self, data, predicted_effects, *args, **kwargs):
        y = data

        mean = self._compute_mean(predicted_effects) * self.scale_
        mean = numpyro.deterministic("mean", mean)
        noise_scale = numpyro.sample("noise_scale", dist.HalfNormal(self.noise_scale))

        if y is not None:
            y = y * self.scale_

        with numpyro.plate("data", len(mean), dim=-2):
            numpyro.sample(
                "obs",
                BetaReparametrized(mean.reshape((-1, 1)), noise_scale),
                obs=y,
            )

        return jnp.zeros_like(mean)



from .optimizer import BudgetOptimizer
from .objectives import (
    MinimizeBudget,
    MaximizeKPI,
    MaximizeROI,
)
from .constraints import (
    TotalBudgetConstraint,
    MinimumTargetResponse,
    SharedBudgetConstraint,
)
from .parametrization_transformations import (
    IdentityTransform,
    InvestmentPerChannelTransform,
    TotalInvestmentTransform,
    InvestmentPerChannelAndSeries,
    InvestmentPerSeries,
)

__all__ = [
    "BudgetOptimizer",
    "MinimizeBudget",
    "MaximizeKPI",
    "MaximizeROI",
    "TotalBudgetConstraint",
    "MinimumTargetResponse",
    "IdentityTransform",
    "InvestmentPerChannelTransform",
    "TotalInvestmentTransform",
    "SharedBudgetConstraint",
    "InvestmentPerChannelAndSeries",
    "InvestmentPerSeries",
]



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from prophetverse.effects.base import BaseEffect


def plot_prior_predictive(
    effect: BaseEffect,
    X=None,
    y=None,
    predicted_effects=None,
    num_samples=1000,
    coverage=0.95,
    matplotlib_kwargs=None,
    mode="time",
):

    if (X is not None) and (X.index.nlevels > 1):  # pragma : no cover
        raise ValueError("This utility does not work with panel data yet.")

    if (y is not None) and (y.index.nlevels > 1):
        raise ValueError("This utility does not work with panel data yet.")

    samples = effect.sample_prior(
        X=X,
        y=y,
        num_samples=num_samples,
        predicted_effects=predicted_effects,
        as_pandas=True,
        seed=42,
    ).unstack(level=0)

    idx = samples.index.get_level_values(-1).unique()

    if isinstance(idx, pd.PeriodIndex):
        idx = idx.to_timestamp()

    alpha = (1 - coverage) / 2
    quantiles = [alpha, 1 - alpha]

    prior_samples = samples.quantile(quantiles, axis=1).T

    matplotlib_kwargs = {} if matplotlib_kwargs is None else matplotlib_kwargs

    fig, ax = plt.subplots(**matplotlib_kwargs)

    if mode == "time":
        _x = idx
        argsort = np.arange(len(idx))
    else:
        _x = X[mode].values.flatten()
        argsort = np.argsort(_x)

    ax.fill_between(
        x=_x[argsort],
        y1=prior_samples[quantiles[0]].values.flatten()[argsort],
        y2=prior_samples[quantiles[1]].values.flatten()[argsort],
        alpha=0.2,
    )

    ax.plot(_x[argsort], samples.mean(axis=1).iloc[argsort])

    return fig, ax



"""Target effects for Prophetverse models."""

from .base import BaseTargetEffect
from .multivariate import MultivariateNormal
from .univariate import (
    NormalTargetLikelihood,
    GammaTargetLikelihood,
    BetaTargetLikelihood,
    NegativeBinomialTargetLikelihood,
)


__all__ = [
    "NormalTargetLikelihood",
    "GammaTargetLikelihood",
    "BetaTargetLikelihood",
    "NegativeBinomialTargetLikelihood",
    "MultivariateNormal",
]



import numpy as np
import pandas as pd


def load_composite_effect_example():
    """
    Load a synthetic time series with a composite effect.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the time series data.
    """
    rng = np.random.default_rng(0)
    timeindex = pd.period_range(
        start="2010-01-01", freq="D", periods=365 * 7, name="time"
    )

    t = np.arange(len(timeindex))

    w = np.ones(100) / 100
    trend = np.ones(len(t)) * t / 20 + 10

    seasonality = (
        np.sin(2 * np.pi * t / 365.25) * 0.7
        + np.sin(2 * np.pi * t / 365.25 * 2) * 1
        # + np.sin(2 * np.pi * t / 365.25 * 3) * 0.5
        # + np.sin(2 * np.pi * t / 365.25 * 4) * 0.5
    ) * 0.8 + 1

    exog = np.clip(rng.normal(0.1, 1, size=len(t)), 0, None)
    # rolling mean
    w = np.ones(15) / 15
    exog = np.convolve(exog, w, mode="same")
    exog -= np.min(exog)
    exog_effect = exog * 0.5
    noise = rng.normal(0, 0.1, size=len(t))
    y = pd.DataFrame(
        data={
            "target": trend * (1 + exog_effect + seasonality + noise)
            + trend * exog * (seasonality - seasonality.min() + 1) * 2
        },
        index=timeindex,
    )

    X = pd.DataFrame(
        {
            "investment": exog,
        },
        index=timeindex,
    )
    return y, X



"""Constant effect module."""

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from numpyro.distributions import Distribution

from prophetverse.effects.base import BaseEffect


class Constant(BaseEffect):
    """Constant effect.

    Implements a constant effect.

    Parameters
    ----------
    prior : Distribution, optional
        The prior distribution for the constant coefficient, by default None
        which corresponds to a standard normal distribution.
    """

    _tags = {
        "requires_X": False,
    }

    def __init__(self, prior: Distribution = None) -> None:
        self.prior = prior
        super().__init__()

        self._prior = prior
        if self._prior is None:
            self._prior = dist.Normal(0, 1)

    def _transform(self, X, fh):
        return jnp.ones((len(fh), 1))

    def _predict(  # type: ignore[override]
        self, data: jnp.ndarray, predicted_effects: dict, *args, **kwargs
    ) -> jnp.ndarray:
        """Apply the trend.

        Parameters
        ----------
        constant_vector : jnp.ndarray
            A constant vector with the size of the series time indexes

        Returns
        -------
        jnp.ndarray
            The forecasted trend
        """
        # Alias for clarity

        coefficient = numpyro.sample("constant_coefficient", self._prior)

        return coefficient * jnp.ones_like(data)



"""ROI (Return on Investment) likelihood effect."""

from typing import Any, Dict

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import pandas as pd

from prophetverse.utils.frame_to_array import series_to_tensor_or_array

from .base import BaseEffect

__all__ = ["ROILikelihood"]


class ROILikelihood(BaseEffect):
    """Apply a normal likelihood to the ROI (Return on Investment) of an effect.

    This effect computes the ROI as the ratio between the sum of the effect's
    contribution and the sum of the input (captured at transform time). A normal
    likelihood is applied to this ratio, centered at a given value with a given scale.

    The ROI is computed as:
        ROI = sum(effect_contribution) / sum(input)

    Parameters
    ----------
    effect_name : str
        The name of the effect to use for computing ROI. This should match
        the key in predicted_effects dictionary.
    roi_mean : float
        The expected (center) value of the ROI. The normal likelihood will be
        centered at this value.
    roi_scale : float
        The scale (standard deviation) of the normal distribution for the ROI.
        Must be greater than 0.

    Examples
    --------
    >>> from prophetverse.effects import ROILikelihood
    >>> roi_effect = ROILikelihood(
    ...     effect_name="marketing_spend",
    ...     roi_mean=2.0,  # Expect $2 return per $1 spent
    ...     roi_scale=0.5,
    ... )
    """

    _tags = {"requires_X": True, "capability:panel": False}

    def __init__(
        self,
        effect_name: str,
        roi_mean: float,
        roi_scale: float,
    ):
        self.effect_name = effect_name
        self.roi_mean = roi_mean
        self.roi_scale = roi_scale

        assert self.roi_scale > 0, "roi_scale must be greater than 0"

        super().__init__()

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1):
        """Initialize the effect.

        This method is called during `fit()` of the forecasting model.
        It receives the Exogenous variables DataFrame and should be used to initialize
        any necessary parameters or data structures.

        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.

        scale : float, optional
            The scale of the timeseries. For multivariate timeseries, this is
            a dataframe. For univariate, it is a simple float.

        Returns
        -------
        None
        """
        self.timeseries_scale = scale

    def _transform(self, X: pd.DataFrame, fh: pd.Index) -> Dict[str, Any]:
        """Prepare input data to be passed to numpyro model.

        Captures the input values that will be used to compute ROI.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables for the training
            time indexes, if passed during fit, or for the forecasting time indexes, if
            passed during predict.

        fh : pd.Index
            The forecasting horizon as a pandas Index.

        Returns
        -------
        Dict[str, Any]
            Dictionary with the captured input data for ROI computation.
        """
        data_dict = {}

        # Capture the input values - sum across all columns if multiple
        input_array = series_to_tensor_or_array(X)
        data_dict["input_values"] = input_array
        data_dict["input_sum"] = jnp.sum(input_array)
        data_dict["data"] = None

        return data_dict

    def _predict(
        self, data: Dict, predicted_effects: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Computes the ROI and applies a normal likelihood centered at the expected
        ROI value.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with zeros (this effect only adds the likelihood constraint).
        """
        input_sum = data["input_sum"]

        # Get the effect contribution
        effect_contribution = predicted_effects[self.effect_name]
        effect_sum = jnp.sum(effect_contribution)

        # Compute ROI as the ratio of effect sum to input sum
        # Add small epsilon to avoid division by zero
        roi = effect_sum / (input_sum + 1e-10)

        # Apply normal likelihood to the ROI
        # Use :ignore suffix so that this sample is removed from output dataframe
        numpyro.sample(
            "roi_likelihood:ignore",
            dist.Normal(self.roi_mean, self.roi_scale),
            obs=roi,
        )

        # Return zeros - this effect only adds a likelihood constraint
        return jnp.zeros_like(effect_contribution)

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return.

        Returns
        -------
        params : list of dict
            Parameters to create testing instances of the class.
        """
        return [
            {
                "effect_name": "trend",
                "roi_mean": 2.0,
                "roi_scale": 0.5,
            }
        ]



"""Dataset for lifttest example."""

import numpy as np
import numpyro.distributions as dist
import pandas as pd
from pathlib import Path
from prophetverse.effects import (
    HillEffect,
    LinearEffect,
    LinearFourierSeasonality,
    ChainedEffects,
    GeometricAdstockEffect,
)
from prophetverse.effects.trend import PiecewiseLinearTrend
from prophetverse.engine.prior import PriorPredictiveInferenceEngine
from prophetverse.experimental.simulate import simulate
from prophetverse.sktime import Prophetverse
from prophetverse.utils.regex import exact, no_input_columns
from prophetverse.effects.target.univariate import NegativeBinomialTargetLikelihood
import jax.numpy as jnp
import json


def get_index():
    """
    Generate a time index ranging from 2000-01-01 to 2005-01-01 with daily frequency.

    Returns
    -------
    pd.PeriodIndex
        The generated time index.
    """
    index = pd.period_range("2000-01-01", "2005-01-01", freq="D")

    index = pd.MultiIndex.from_product([["a", "b"], index], names=["group", "date"])
    return index


def get_X(index, rng):
    """
    Create a DataFrame of two simulated investments with daily data.

    Parameters
    ----------
    index : pd.PeriodIndex
        The time index for the DataFrame.

    Returns
    -------
    pd.DataFrame
        DataFrame containing normalized simulated investments.
    """

    X = pd.DataFrame(
        {
            "ad_spend_search": np.cumsum(rng.normal(0, 1, size=len(index))),
        },
        index=index,
    )
    X -= X.min()
    X /= X.max()
    X += 0.05

    X["ad_spend_social_media"] = (
        X["ad_spend_search"] + 0.1 + rng.normal(0, 0.01, size=len(index))
    )
    X["ad_spend_social_media"] = X["ad_spend_social_media"] ** 3.5
    X *= 100_000

    return X


def get_site_values():

    posterior_samples_path = Path(__file__).parent / Path(
        "dataset1_panel_posterior_samples.json"
    )
    with open(posterior_samples_path, "r") as f:
        posterior_samples = json.load(f)

    for key in posterior_samples.keys():
        posterior_samples[key] = jnp.array(posterior_samples[key])

    return posterior_samples


def get_groundtruth_model():
    """
    Define and configure a Prophetverse model with custom components.

    Returns
    -------
    Prophetverse
        Configured Prophetverse model.
    """

    site_samples = get_site_values()
    model = Prophetverse(
        trend=PiecewiseLinearTrend(
            changepoint_interval=100,
            changepoint_range=-100,
            remove_seasonality_before_suggesting_initial_vals=False,
        ),
        exogenous_effects=[
            (
                "yearly_seasonality",
                LinearFourierSeasonality(
                    freq="D",
                    sp_list=[365.25],
                    fourier_terms_list=[3],
                    prior_scale=0.05,
                    effect_mode="multiplicative",
                ),
                no_input_columns,
            ),
            (
                "weekly_seasonality",
                LinearFourierSeasonality(
                    freq="D",
                    sp_list=[7],
                    fourier_terms_list=[3],
                    prior_scale=0.01,
                    effect_mode="multiplicative",
                ),
                no_input_columns,
            ),
            (
                "monthly_seasonality",
                LinearFourierSeasonality(
                    freq="D",
                    sp_list=[28],
                    fourier_terms_list=[5],
                    prior_scale=0.05,
                    effect_mode="multiplicative",
                ),
                no_input_columns,
            ),
            (
                "ad_spend_search",
                HillEffect(
                    effect_mode="additive",
                ),
                exact("ad_spend_search"),
            ),
            (
                "ad_spend_social_media",
                ChainedEffects(
                    steps=[
                        ("adstock", GeometricAdstockEffect()),
                        ("saturation", HillEffect(effect_mode="additive")),
                    ]
                ),
                exact("ad_spend_social_media"),
            ),
        ],
        inference_engine=PriorPredictiveInferenceEngine(
            num_samples=1, substitute=site_samples
        ),
        # scale=1,
        broadcast_mode="effect",
    )

    return model


def get_samples(model, X):
    """
    Simulate samples from the provided model and input data.

    Parameters
    ----------
    model : Prophetverse
        The Prophetverse model to use for simulation.
    X : pd.DataFrame
        Exogenous data for the simulation.

    Returns
    -------
    dict
        Simulated samples from the model.
    """
    samples, model = simulate(model=model, fh=X.index, X=X, return_model=True)
    return samples, model


def get_y(samples, index):
    """
    Extract observed sales data from the simulated samples.

    Parameters
    ----------
    samples : dict
        Simulated samples from the model.
    index : pd.PeriodIndex
        Time index for the observed data.

    Returns
    -------
    pd.DataFrame
        DataFrame containing observed sales data.
    """
    mask = samples.index.get_level_values("sample") == 0
    return samples.loc[mask, "obs"].to_frame("sales")


def get_simulated_lift_test(X, model, true_effect, rng, n=10):
    """
    Perform a simulated lift test by perturbing exogenous variables.

    Parameters
    ----------
    X : pd.DataFrame
        Original exogenous data.
    model : Prophetverse
        The Prophetverse model to use for simulation.
    samples : dict
        Simulated samples from the model.
    true_effect : pd.DataFrame
        True effects of the exogenous variables.
    n : int, optional
        Number of samples to return for the lift test, by default 10.

    Returns
    -------
    tuple of pd.DataFrame
        Lift test results for each exogenous variable.
    """

    outs = []
    fh = X.index.get_level_values(-1).unique()
    for col in ["ad_spend_search", "ad_spend_social_media"]:

        X_b = X.copy()

        X_b[col] = X_b[col] * rng.uniform(0.8, 1.5, size=X.shape[0])

        samples_b = model.predict_component_samples(X=X_b, fh=fh)

        mask = samples_b.index.get_level_values("sample") == 0
        true_effect_b = samples_b.loc[mask, [col]].droplevel("sample")

        lift = true_effect_b / true_effect

        lift_test_dataframe = pd.DataFrame(
            index=X.index,
            data={
                "lift": (lift[col] * rng.normal(1, 0.05, size=X.shape[0])),
                "x_start": X.loc[:, col],
                "x_end": X_b.loc[:, col],
            },
        )
        outs.append(lift_test_dataframe.sample(n=n, replace=False, random_state=42))

    return tuple(outs)


def get_dataset():
    """
    Generate a complete dataset.

    Includes time index, exogenous data, model,
    simulated samples, observed sales, true effects, and lift test results.

    Returns
    -------
    tuple
        Contains observed sales, exogenous data, lift test results, true effects,
        and the Prophetverse model.
    """

    rng = np.random.default_rng(0)
    index = get_index()
    X = get_X(index, rng=rng)
    model = get_groundtruth_model()

    fh = index.get_level_values(-1).unique()
    model.fit(X=X, y=pd.DataFrame(np.zeros(X.shape[0]), index=X.index))
    y = model.predict(X=X, fh=fh)
    true_effect = model.predict_components(X=X, fh=fh)
    lift_test = get_simulated_lift_test(X, model, true_effect, n=30, rng=rng)

    return y, X, lift_test, true_effect, model



"""Base classes for sktime forecasters in prophetverse."""

import itertools
import warnings
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist
import pandas as pd
from sktime.base import _HeterogenousMetaEstimator
from sktime.forecasting.base import BaseForecaster, ForecastingHorizon

from prophetverse._model import model as model_func
from prophetverse.effects.base import BaseEffect
from prophetverse.effects.linear import LinearEffect
from prophetverse.effects.trend import (
    FlatTrend,
    PiecewiseLinearTrend,
    PiecewiseLogisticTrend,
)
from prophetverse.engine import (
    BaseInferenceEngine,
    MAPInferenceEngine,
    MCMCInferenceEngine,
)
from prophetverse.engine.optimizer.optimizer import CosineScheduleAdamOptimizer
from prophetverse.utils import get_multiindex_loc

_VALID_TREND_STRINGS = ["linear", "logistic", "flat"]


class BaseBayesianForecaster(BaseForecaster):
    """

    A base class for numpyro Bayesian forecasters in sktime.

    Specifies methods and signatures that all Bayesian forecasters have to implement,
    and handles optimization and sampling of the posterior distribution.


    Parameters
    ----------
    rng_key: KeyArray
        The RNG Key to use for sampling.
    scale: Union[float, pd.Series, pd.DataFrame]
        The scaling factor for the target variable.
    inference_engine: Optional[BaseInferenceEngine]
        The inference engine to use for sampling.
        If None, the default MCMCInferenceEngine is used.
    """

    _tags = {
        "capability:pred_int": True,
        "capability:pred_int:insample": True,
        "enforce_index_type": [pd.Period, pd.DatetimeIndex],
    }

    def __init__(
        self,
        rng_key: jax.typing.ArrayLike = None,
        scale=None,
        inference_engine: Optional[BaseInferenceEngine] = None,
    ):

        self.rng_key = rng_key
        self.scale = scale
        self.inference_engine = inference_engine
        super().__init__()

        if self.scale:
            self._scale = self.scale
        else:
            self._scale = None

        if self.inference_engine is not None:
            self._inference_engine = self.inference_engine
        else:
            self._inference_engine = MCMCInferenceEngine()

    @property
    def _likelihood_is_discrete(self):
        """Property that indicates whether the forecaster uses a discrete likelihood.

        As a consequence, the target variable must be integer-valued and will not be
        scaled before _get_fit_data is called.

        Returns
        -------
        bool
            True if the forecaster uses a discrete likelihood, False otherwise.
        """
        return False

    @property
    def posterior_samples_(self) -> Dict[str, np.ndarray]:
        """Property returning the samples of the inference engine.

        Returns
        -------
        dict
            Dictionary containing samples with trace names as key.
        """
        return self.inference_engine_.posterior_samples_

    # pragma: no cover
    def _get_fit_data(
        self, y: pd.DataFrame, X: pd.DataFrame, fh: ForecastingHorizon
    ) -> Dict[str, Any]:
        """
        Get the data required for the Numpyro model.

        Parameters
        ----------
        y : pd.DataFrame
            Target variable.
        X : pd.DataFrame
            Exogenous variables.
        fh : ForecastingHorizon
            Forecasting horizon.

        Returns
        -------
        Dict[str, Any]
            Data required for the Numpyro model.

        Raises
        ------
        NotImplementedError
            If the method is not implemented by the subclass.
        """
        raise NotImplementedError("Must be implemented by subclass")

    def get_predict_data(self, X: pd.DataFrame, fh: ForecastingHorizon):
        if not isinstance(fh, ForecastingHorizon):
            fh = self._check_fh(fh)
        X = self._check_X(X=X)
        return self._get_predict_data(X=X, fh=fh)

    # pragma: no cover
    def _get_predict_data(self, X: pd.DataFrame, fh: ForecastingHorizon):
        """Generate samples from the posterior predictive distribution.

        Parameters
        ----------
        X : pd.DataFrame
            Exogenous variables.
        fh : ForecastingHorizon
            Forecasting horizon.

        Returns
        -------
        pd.DataFrame
            Samples from the posterior predictive distribution.
        """
        raise NotImplementedError("Must be implemented by subclass")

    def model(self, *args, **kwargs):
        """Numpyro model. If not overridden by subclass, defaults to :func:`model_func`.

        Parameters
        ----------
        *args : tuple
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Any
            Model output.
        """
        return model_func(*args, **kwargs)

    def _fit(self, y, X, fh):
        """
        Fit the Bayesian forecaster to the training data.

        Parameters
        ----------
        y : pd.DataFrame
            Target variable.
        X : pd.DataFrame
            Exogenous variables.
        fh : ForecastingHorizon
            Forecasting horizon.

        Returns
        -------
        self : object
            The fitted Bayesian forecaster.
        """
        self._set_y_scales(y)
        y = self._scale_y(y)

        self.internal_y_indexes_ = y.index
        data = self._get_fit_data(y, X, fh)

        self.distributions_ = data.get("distributions", {})

        self.inference_engine_ = self._inference_engine.clone()
        self.inference_engine_.infer(self.model, **data)

        return self

    def _predict(self, fh, X):
        """
        Generate point forecasts for the given forecasting horizon.

        Parameters
        ----------
        fh : ForecastingHorizon
            Forecasting horizon.
        X : pd.DataFrame
            Exogenous variables.

        Returns
        -------
        pd.DataFrame
            Point forecasts for the forecasting horizon.
        """
        predictive_samples = self.predict_components(fh=fh, X=X)
        mean = predictive_samples["mean"]

        # TODO: can `feature_names` be empty?
        target_names = self._y_metadata["feature_names"]
        y_pred = mean.to_frame(target_names[0])

        return self._postprocess_output(y_pred)

    def predict_components(self, fh: ForecastingHorizon, X: pd.DataFrame = None):
        """
        Predicts the values for all sites.

        Given a forecast horizon and optional input features, returns a DataFrame
        the mean of the predicted values for all sites.

        Parameters
        ----------
        fh : ForecastingHorizon
            The forecast horizon, specifying the number of time steps to predict into
            the future.
        X : array-like, optional
            The input features used for prediction. Defaults to None.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the predicted values for all sites, with the site
            names as columns and the forecast horizon as the index.
        """
        if self._is_vectorized:
            return self._vectorize_predict_method("predict_components", X=X, fh=fh)

        fh_as_index = self.fh_to_index(fh)

        X_inner = self._check_X(X=X)
        predictive_samples_ = self._get_predictive_samples_dict(fh=fh, X=X_inner)

        out = pd.DataFrame(
            data={
                site: data.mean(axis=0).flatten()
                for site, data in predictive_samples_.items()
            },
            index=self.periodindex_to_multiindex(fh_as_index),
        ).sort_index()

        return self._inv_scale_y(out)

    def _postprocess_output(self, y: pd.DataFrame) -> pd.DataFrame:
        """Postprocess outputs. Default to do nothing.

        This method is a placeholder for child classes that may need
        to do specific postprocesing, for example the HierarchicalProphet

        Parameters
        ----------
        y : pd.DataFrame
            dataframe with output predictions

        Returns
        -------
        pd.DataFrame
            postprocessed dataframe

        """
        return y

    def _get_predictive_samples_dict(
        self, fh: ForecastingHorizon, X: Optional[pd.DataFrame] = None
    ) -> dict[str, jnp.ndarray]:
        """
        Return a dictionary of predictive samples for each time series.

        Parameters
        ----------
        fh : ForecastingHorizon or int
            The forecasting horizon specifying the time points to forecast.
        X : array-like, optional (default=None)
            The input features for the time series forecasting model.

        Returns
        -------
        predictive_samples_dict : dict[str, jnp.ndarray]
            A dictionary containing predictive samples for each time series in the input
            data. The keys are the names of the time series, and the values are NumPy
            arrays representing the predictive samples.
        """

        predict_data = self.get_predict_data(X=X, fh=fh)

        predictive_samples_ = self.inference_engine_.predict(**predict_data)

        keys_to_delete = []
        for key in predictive_samples_.keys():
            if ":ignore" in key:
                keys_to_delete.append(key)
        for key in keys_to_delete:
            del predictive_samples_[key]
        return predictive_samples_

    def predict_component_samples(self, fh, X=None):
        """
        Predicts samples for all sites.

        Parameters
        ----------
        fh : int or array-like
            The forecast horizon or an array-like object representing the forecast
            horizons.
        X : array-like, optional
            The input features for prediction. Defaults to None.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the predicted samples for all sites.
        """
        if self._is_vectorized:
            out = self._vectorize_predict_method(
                "predict_component_samples", X=X, fh=fh
            )

            return out

        predictive_samples_ = self._get_predictive_samples_dict(fh=fh, X=X)

        fh_as_index = self.fh_to_index(fh)
        dfs = []
        for site, data in predictive_samples_.items():

            idxs = self.periodindex_to_multiindex(fh_as_index)

            samples_idx = np.arange(data.shape[0])

            def _coerce_to_tuple(x):
                if isinstance(x, tuple):
                    return x
                return (x,)

            tuples = [
                (sample_i, *_coerce_to_tuple(idx))
                for sample_i, idx in itertools.product(samples_idx, idxs)
            ]
            # Set samples_idx as level 0 of idx
            idx = pd.MultiIndex.from_tuples(tuples, names=["sample", *idxs.names])

            flattened_data = data.flatten()

            if flattened_data.shape[0] != len(idx):
                warnings.warn(
                    f"Data shape {flattened_data.shape} does not match index shape {len(idx)}"
                    f"for site {site}. This site will not be returned."
                )
                continue
            dfs.append(
                pd.DataFrame(
                    data={site: data.flatten()},
                    index=idx,
                )
            )

        df = pd.concat(dfs, axis=1)

        # Scale but apply the transformation to each sample at index level 0
        df = df.groupby(level=0).apply(lambda x: self._inv_scale_y(x.droplevel(0)))
        df = df.sort_index(level=0)

        return df

    def predict_samples(self, fh: ForecastingHorizon, X: Optional[pd.DataFrame] = None):
        """
        Generate samples from the posterior predictive distribution.

        Parameters
        ----------
        X : pd.DataFrame, optional
            Exogenous variables.
        fh : ForecastingHorizon
            Forecasting horizon.

        Returns
        -------
        pd.DataFrame
            Samples from the posterior predictive distribution.
        """
        if self._is_vectorized:
            return self._vectorize_predict_method("predict_samples", X=X, fh=fh)

        fh_as_index = self.fh_to_index(fh)

        X_inner = self._check_X(X=X)
        predictive_samples_ = self._get_predictive_samples_dict(fh=fh, X=X_inner)

        observation_site = predictive_samples_["obs"]
        n_samples = predictive_samples_["obs"].shape[0]
        preds = pd.DataFrame(
            data=observation_site.T.reshape((-1, n_samples)),
            columns=list(range(n_samples)),
            index=self.periodindex_to_multiindex(fh_as_index),
        ).sort_index()

        return self._inv_scale_y(preds)

    def _set_y_scales(self, y: pd.DataFrame):
        """
        Set the scaling factor for the target variable.

        * If `scale` attribute is set, it will be used as the scaling factor.
        * If the target variable has only one level, the scaling factor will be set
            as the maximum absolute value of the target variable.
        * If the target variable has multiple levels, the scaling factor will be
            set as the maximum absolute value of each level, grouped by the remaining
            levels.


        Parameters
        ----------
            y (pd.Series or pd.DataFrame): The target variable.

        Returns
        -------
            None
        """
        if self.scale is not None:
            self._scale = self.scale
        elif y.index.nlevels == 1:
            self._scale = float(y.abs().max().values[0])
        else:
            self._scale = y.groupby(level=list(range(y.index.nlevels - 1))).agg(
                lambda x: np.abs(x).max()
            )

        if isinstance(self._scale, (float, int)):
            if self._scale == 0:
                self._scale = 1
        elif isinstance(self._scale, (pd.Series, pd.DataFrame)):
            # Map any values that are 0 to 1
            self._scale = self._scale.replace(0, 1)

    def _scale_y(self, y: pd.DataFrame, force=False) -> pd.DataFrame:
        """
        Scales the input DataFrame y (divide it by the scaling factor).

        Parameters
        ----------
        y : pd.DataFrame
            The input DataFrame to be inverse scaled.
        force: bool, optional
            Should scale even if likelihood is discrete.

        Returns
        -------
        pd.DataFrame
            The inverse scaled DataFrame.

        Notes
        -----
        This method takes a DataFrame `y` as input and performs scaling on it.
        If scaling is skipped, the original DataFrame is returned.
        If the scaling factor is a float, each value in the DataFrame is divided by the
        scaling factor.
        If the scaling factor is a DataFrame, the scaling factor for each observation is
        determined based on the index of the input DataFrame `y`.
        The input DataFrame `y` is then divided by the corresponding scaling factor for
        each observation.
        This method assumes that the scaling factor has already been computed and stored
        in the `_scale` attribute of the class.
        """
        if self._likelihood_is_discrete and not force:
            return y

        if isinstance(self._scale, (int, float)):
            return y / self._scale

        scale_for_each_obs = self._scale.loc[y.index.droplevel(-1)].values
        return y / scale_for_each_obs

    def _inv_scale_y(self, y: pd.DataFrame) -> pd.DataFrame:
        """
        Inverse scales the input DataFrame y.

        Parameters
        ----------
        y : pd.DataFrame
            The input DataFrame to be inverse scaled.

        Returns
        -------
        pd.DataFrame
            The inverse scaled DataFrame.

        Notes
        -----
        This method takes a DataFrame `y` as input and performs inverse scaling on it.
        If scaling is skipped, the original DataFrame is returned.
        If the scaling factor is a float, each value in the DataFrame is multiplied by
        the scaling factor.
        If the scaling factor is a DataFrame, the scaling factor for each observation
        is determined based on the index of the input DataFrame `y`.
        The input DataFrame `y` is then multiplied by the corresponding scaling factor
        for each observation.
        This method assumes that the scaling factor has already been computed and stored
        in the `_scale` attribute of the class.
        """
        if self._likelihood_is_discrete:
            return y

        if isinstance(self._scale, (float, int)):
            return y * self._scale
        scale_for_each_obs = self._scale.loc[y.index.droplevel(-1)].values
        return y * scale_for_each_obs

    def _predict_quantiles(
        self,
        fh: ForecastingHorizon,
        X: Optional[pd.DataFrame],
        alpha: Union[float, List[float]],
    ):
        """
        Generate quantile forecasts for the given forecasting horizon.

        Parameters
        ----------
        fh: ForecastingHorizon
            The Forecasting horizon.
        X: pd.DataFrame
            Exogenous variables dataframe
        alpha: float or List[float])
            Quantile(s) to compute.

        Returns
        -------
        pd.DataFrame
            Quantile forecasts for the forecasting horizon.
        """
        if isinstance(alpha, float):
            alpha = [alpha]

        forecast_samples_ = self.predict_samples(X=X, fh=fh)

        var_names = self._get_varnames()

        int_idx = pd.MultiIndex.from_product([var_names, alpha])

        quantiles = forecast_samples_.quantile(alpha, axis=1).T
        quantiles.columns = int_idx

        return quantiles

    def periodindex_to_multiindex(self, periodindex: pd.PeriodIndex) -> pd.MultiIndex:
        """
        Convert a PeriodIndex to a MultiIndex.

        Parameters
        ----------
        periodindex: pd.PeriodIndex
            PeriodIndex to convert.

        Returns
        -------
        pd.MultiIndex
            Converted MultiIndex.
        """
        if self._y.index.nlevels == 1:
            return periodindex

        series_id_tuples = self.internal_y_indexes_.droplevel(-1).unique().tolist()

        # Check if base_levels 0 is a iterable:
        if not isinstance(series_id_tuples[0], tuple):
            series_id_tuples = [(x,) for x in series_id_tuples]

        series_id_tuples = self._filter_series_tuples(series_id_tuples)

        return pd.Index(
            map(
                lambda x: (*x[0], x[1]),
                itertools.product(series_id_tuples, periodindex),
            ),
            name=self.internal_y_indexes_.names,
        )

    def _filter_series_tuples(self, levels: List[Tuple]) -> List[Tuple]:
        """Filter series levels.

        This method can be overriden by subclasses to filter the series levels.
        See the multivariate model for an example.

        Parameters
        ----------
        levels : List[Tuple]
            The multiindex levels of the dataframe (excluding the last level, which is
            the time index).

        Returns
        -------
        List[Tuple]
            The multiindex levels of the dataframe after filtering.

        """
        return levels

    def fh_to_index(self, fh: ForecastingHorizon, y: pd.DataFrame = None):
        """
        Convert a ForecastingHorizon to an index.

        Parameters
        ----------
        fh : ForecastingHorizon
            Forecasting horizon.
        y : pd.DataFrame, optional
            Target variable. Defaults to None.

        Returns
        -------
        pd.Index
            Index corresponding to the forecasting horizon.
        """
        if not isinstance(fh, ForecastingHorizon):
            fh = self._check_fh(fh)

        if y is None:
            y = self._y
        fh_dates = fh.to_absolute_index(cutoff=y.index.get_level_values(-1).max())
        return fh_dates

    @property
    def var_names(self):
        """
        Return variable names.

        Returns
        -------
        List[str]
            List of variable names.
        """
        return list(self.distributions_.keys())

    def plot_trace(self, *, figsize=(15, 25), compact=True, var_names=None, **kwargs):
        """
        Plot trace of the MCMC samples using Arviz.

        Parameters
        ----------
        figsize : tuple, optional
            Figure size. Default is (15, 25).
        compact : bool, optional
            Whether to use compact display. Default is True.
        var_names : list, optional
            List of variable names to plot. Default is None.

        Returns
        -------
        Figure
            Arviz plot trace figure.
        """
        import arviz as az

        if var_names is None:
            var_names = self.var_names

        return az.plot_trace(
            az.from_numpyro(self.inference_engine_.mcmc_),
            var_names=var_names,
            filter_vars="regex",
            compact=compact,
            figsize=figsize,
        )

    @property
    def site_names(self) -> list[Any]:
        """
        Return site names.

        Returns
        -------
        List[Any]
            List of site names.
        """
        return list(self.posterior_samples_.keys())

    def _vectorize_predict_method(
        self, methodname: str, X: pd.DataFrame, fh: ForecastingHorizon
    ):
        """
        Handle sktime's "vectorization" of timeseries.

        When a multiindex is passed as input and the forecaster does not accept natively
        hierarchical data, it will create a dataframe of forecasters, and they should be
        called with the corresponding index.

        Parameters
        ----------
        methodname : str
            The method name to call.
        X : pd.DataFrame
            The input data.
        fh : ForecastingHorizon
            The forecasting horizon.

        Returns
        -------
        pd.DataFrame
            The output of the method.
        """
        if not self._is_vectorized:
            return getattr(self, methodname)(X=X, fh=fh)

        def _coerce_to_tuple(x):
            if isinstance(x, (tuple, list)):
                return x
            return (x,)

        outs = []
        for idx, data in self.forecasters_.iterrows():
            forecaster = data[0]

            if X is None:
                _X = None
            else:
                _X = get_multiindex_loc(X, [idx])
                # Keep only index level -1
                for _ in range(_X.index.nlevels - 1):
                    _X = _X.droplevel(0)
            out = getattr(forecaster, methodname)(X=_X, fh=fh)

            if not isinstance(idx, (tuple, list)):
                idx = [idx]  # pragma: no cover
            new_index = pd.MultiIndex.from_tuples(
                [[*idx, *_coerce_to_tuple(dateidx)] for dateidx in out.index],
                names=[*self.forecasters_.index.names, *out.index.names],
            )
            out.set_index(new_index, inplace=True)
            outs.append(out)
        return pd.concat(outs, axis=0)


class BaseProphetForecaster(_HeterogenousMetaEstimator, BaseBayesianForecaster):
    """Base class for Bayesian estimators with Effects objects.

    Parameters
    ----------
    trend : Union[str, BaseEffect], optional
        One of "linear" (default), "linear1" or "logistic". Type of trend to use.
        Can also be a custom effect object.

    changepoint_interval : int, optional, default=25
        Number of potential changepoints to sample in the history.

    changepoint_range : float or int, optional, default=0.8
        Proportion of the history in which trend changepoints will be estimated.

        * if float, must be between 0 and 1.
          The range will be that proportion of the training history.

        * if int, can be positive or negative.
          Absolute value must be less than number of training points.
          The range will be that number of points.
          A negative int indicates number of points
          counting from the end of the history, a positive int from the beginning.

    changepoint_prior_scale : float, optional, default=0.001
        Regularization parameter controlling the flexibility
        of the automatic changepoint selection.

    offset_prior_scale : float, optional, default=0.1
        Scale parameter for the prior distribution of the offset.
        The offset is the constant term in the piecewise trend equation.

    capacity_prior_scale : float, optional, default=0.2
        Scale parameter for the prior distribution of the capacity.

    capacity_prior_loc : float, optional, default=1.1
        Location parameter for the prior distribution of the capacity.

    exogenous_effects : List[Tuple[str, BaseEffect, str]]
        List of exogenous effects to apply to the data. Each item of the list
        is a tuple with the name of the effect, the effect object, and the regex
        pattern to match the columns of the dataframe.

    default_effect : Optional[BaseEffect]
        Default effect to apply to the columns that do not match any regex pattern.
        If None, a LinearEffect is used.
    """

    _steps_attr = "_exogenous_effects"
    _steps_fitted_attr = "exogenous_effects_"

    def __init__(
        self,
        trend: Union[BaseEffect, str] = "linear",
        exogenous_effects: Optional[List[BaseEffect]] = None,
        default_effect: Optional[BaseEffect] = None,
        rng_key: jax.typing.ArrayLike = None,
        inference_engine: Optional[BaseInferenceEngine] = None,
        scale=None,
    ):

        # Trend related hyperparams
        self.trend = trend

        # Exogenous variables related hyperparams
        self.exogenous_effects = exogenous_effects
        self.default_effect = default_effect
        super().__init__(
            rng_key=rng_key,
            scale=scale,
            inference_engine=inference_engine,
        )

        self._trend = self._get_trend_model()

    @property
    def _exogenous_effects(self):
        """Return exogenous effects.

        This property is for compatibility with _HeterogenousMetaEstimator.
        """
        if self.exogenous_effects is None:
            return []
        return [(name, effect) for name, effect, _ in self.exogenous_effects]

    @_exogenous_effects.setter
    def _exogenous_effects(self, value):
        """Set exogenous effects.

        This property is for compatibility with _HeterogenousMetaEstimator.
        """
        # Ensure that user is passing list of (name, effect) or
        # (name, effect, regex) tuples
        assert len(value[0]) in [2, 3], "Invalid value for exogenous_effects"
        len_values = np.all([len(x) == len(value[0]) for x in value])
        assert len_values, "All tuples must have the same length"

        self.exogenous_effects = [
            (name, effect, regex)
            for ((name, effect), (_, _, regex)) in zip(value, self.exogenous_effects)
        ]

    def _fit_effects(
        self, X: Union[None, pd.DataFrame], y: Optional[pd.DataFrame] = None
    ):
        """
        Set custom effects for the features.

        Parameters
        ----------
        feature_names : pd.Index
            List of feature names (obtained with X.columns)
        """
        fitted_effects_list_: List[Tuple[str, BaseEffect, List[str]]] = []
        columns_with_effects: set[str] = set()
        exogenous_effects: Union[List[Tuple[str, BaseEffect, str]], List] = (
            self.exogenous_effects or []
        )

        for effect_name, effect, regex in exogenous_effects:

            if X is not None:
                columns = self._match_columns(X.columns, regex)
                X_columns = X[columns]
            else:
                X_columns = None

            effect = effect.clone()

            effect.fit(  # type: ignore[attr-defined]
                X=X_columns, y=y, scale=self._scale
            )

            if columns_with_effects.intersection(columns):
                msg = "Columns {} are already set".format(
                    columns_with_effects.intersection(columns)
                )

                warnings.warn(msg, UserWarning, stacklevel=2)

            if not len(columns) and not (regex is None or regex == "^$"):
                msg = f"No columns match the regex {regex}"
                warnings.warn(msg, UserWarning, stacklevel=2)

            columns_with_effects = columns_with_effects.union(columns)
            fitted_effects_list_.append((effect_name, effect, columns))

        if X is not None:

            features_without_effects: List[str] = X.columns.difference(
                columns_with_effects
            ).tolist()

            if len(features_without_effects) > 0:
                X_columns = X[features_without_effects]

                if self.default_effect is None:
                    default_effect = LinearEffect(
                        prior=dist.Normal(0, 1),
                        effect_mode="additive",
                    )
                else:
                    default_effect = self.default_effect

                default_effect = default_effect.clone()

                _y = y
                _scale = self._scale
                if self._likelihood_is_discrete:
                    _y = self._scale_y(y, force=True)
                    _scale = 1

                default_effect.fit(
                    X=X[features_without_effects],
                    y=_y,
                    scale=_scale,  # type: ignore[attr-defined]
                )
                fitted_effects_list_.append(
                    (
                        "exogenous_variables_effect",
                        default_effect,
                        features_without_effects,
                    )
                )

        self.exogenous_effects_ = fitted_effects_list_

    def _transform_effects(self, X: pd.DataFrame, fh: pd.Index, y=None) -> OrderedDict:
        """
        Get exogenous data array.

        Parameters
        ----------
        X : pd.DataFrame
            Input data.
        fh : pd.Index
            Forecasting horizon as an index.
        y : pd.DataFrame, optional
            Target variable. Defaults to None.

        Returns
        -------
        dict
            Dictionary of exogenous data arrays.
        """
        out = OrderedDict()
        for effect_name, effect, columns in self.exogenous_effects_:
            # If no columns are found, skip
            if columns is None or len(columns) == 0:
                if effect.get_tag("requires_X"):
                    continue
            if effect.get_tag("applies_to") == "X":
                _X = X[columns]
            elif effect.get_tag("applies_to") == "y":
                _X = y.copy() if y is not None else None

            data: Dict[str, jnp.ndarray] = effect.transform(_X, fh=fh)
            out[effect_name] = data

        return out

    @property
    def non_skipped_exogenous_effect(self) -> dict[str, BaseEffect]:
        """
        Return exogenous effect dictionary.

        Returns
        -------
        dict
            Dictionary of exogenous effects.
        """
        return {
            effect_name: effect
            for effect_name, effect, columns in self.exogenous_effects_
            if len(columns) > 0 or not effect.get_tag("requires_X")
        }

    def _get_trend_model(self):
        """
        Return the trend model based on the specified trend parameter.

        Returns
        -------
        BaseEffect
            The trend model based on the specified trend parameter.

        Raises
        ------
        ValueError
            If the trend parameter is not one of 'linear', 'logistic', 'flat'
            or a BaseEffect instance.
        """
        if isinstance(self.trend, str):
            if self.trend == "linear":
                return PiecewiseLinearTrend()
            elif self.trend == "logistic":
                return PiecewiseLogisticTrend()
            # Raise error because self.trend str is deprecated since 0.6.0
            elif self.trend == "flat":
                return FlatTrend()

        if isinstance(self.trend, BaseEffect):
            return self.trend
        raise ValueError(f"trend must be in {_VALID_TREND_STRINGS} or be a Effect")

    def _validate_hyperparams(self):
        """Validate the hyperparameters."""

        if self.trend not in _VALID_TREND_STRINGS and not isinstance(
            self.trend, BaseEffect
        ):
            raise ValueError(
                f"trend must be in {_VALID_TREND_STRINGS} or be a Effect"  # noqa
            )

    def _match_columns(
        self, columns: Union[pd.Index, List[str]], regex: Union[str, None]
    ) -> pd.Index:
        """Match the columns of the DataFrame with the regex pattern.

        Parameters
        ----------
        columns : pd.Index
            Columns of the dataframe.

        Returns
        -------
        pd.Index
            The columns that match the regex pattern.

        Raises
        ------
        ValueError
            Indicates the abscence of required regex pattern.
        """
        if isinstance(columns, List):
            columns = pd.Index(columns)

        if regex is None:
            return []

        columns = columns.astype(str)
        return columns[columns.str.match(regex)]

    def __rshift__(self, other):
        """Right shift operator.

        This method allows to chain effects and inference engines using the right shift

        Paremeters
        ----------
        other : BaseEffect, tuple, list or BaseInferenceEngine
            The object to chain with the current object.

        Returns
        -------
        BaseProphetForecaster
            A new instance of the class with the specified effect or inference engine.

        Raises
        ------
        ValueError
            If the type of the object is not valid.
        """
        exogenous_effects = (
            [] if self.exogenous_effects is None else self.exogenous_effects
        )
        if isinstance(other, BaseEffect):
            new_obj = self.clone()
            return new_obj.set_params(trend=other)
        if isinstance(other, tuple):

            assert len(other) == 3
            assert isinstance(other[1], BaseEffect)

            new_obj = self.clone()
            return new_obj.set_params(exogenous_effects=[*exogenous_effects, other])

        if isinstance(other, list):
            new_obj = self.clone()
            return new_obj.set_params(exogenous_effects=[*exogenous_effects, *other])

        if isinstance(other, BaseInferenceEngine):
            new_obj = self.clone()
            return new_obj.set_params(inference_engine=other)

        raise ValueError("Invalid type for right shift operator")

    def _sk_visual_block_(self):
        """
        Implement visual block.

        The try except is used since the methods used inside
        _get_default_visual_block are private in sktime, and not public,
        and may be changed in the future.
        """
        try:
            # Tries to get the default visual block
            return self._get_default_visual_block()

        except Exception as e:
            return super()._sk_visual_block_()

    def _get_default_visual_block(self):
        """Make default visual block."""
        from sktime.utils._estimator_html_repr import _get_visual_block, _VisualBlock

        visual_blocks = []
        visual_block_names = []

        steps = getattr(self, self._steps_attr)
        steps = [("trend", self._trend)] + steps
        names, estimators = zip(*steps)

        name_details = [str(est) for est in estimators]

        visual_blocks.append(
            _VisualBlock(
                "serial",
                estimators,
                names=names,
                name_details=name_details,
                dash_wrapped=False,
            )
        )
        visual_block_names.append("effects")

        inference_engine = self._inference_engine
        visual_blocks.append(_get_visual_block(inference_engine))
        visual_block_names.append("inference_engine")

        return _VisualBlock(
            "parallel",
            visual_blocks,
            names=visual_block_names,
            name_details=[None] * len(visual_blocks),
            dash_wrapped=False,
        )



"""Base classes for budget optimization"""

from skbase.base import BaseObject
import pandas as pd
import jax.numpy as jnp


class BaseOptimizationObjective(BaseObject):
    """
    Defines the optimization objective function.

    The __call__ method should be implemented to return the objective function
    that will be optimized.
    """

    _tags = {"backend": None, "name": None}

    def _objective(self, x: jnp.ndarray, budget_optimizer):

        raise NotImplementedError(
            "Utility function must be callable with model, X, period and columns"
        )

    def __call__(self, model, X, horizon, columns):
        """
        Get optimization objective function


        Parameters
        ----------
        model : Prophetverse
            Prophetverse model
        X : pd.DataFrame
            Input data
        horizon : pd.Index
            Forecast horizon
        columns : list
            List of columns to optimize

        Returns
        -------
        objective : callable
            Objective function
        """

        return self._objective


class BaseConstraint(BaseObject):
    """
    Base constraint class.
    Defines the constraint function that will be used in the optimization.
    The __call__ method should be implemented to return a dictionary
    that will be passed to scipy.minimize contraint argument.
    """

    _tags = {"backend": None, "name": None}

    def __call__(self, model, X, horizon, columns):
        """
        Callable constraint function.
        It is expected to be overridden in subclasses.
        Parameters
        ---------
        model : Prophetverse
            Prophetverse model
        X : pd.DataFrame
            Input data
        horizon : pd.Index
            Forecast horizon
        columns : list
            List of columns to optimize

        Returns
        -------
        constraint : dict
            Dictionary with the constraint function and its jacobian, and type.
        """

        raise NotImplementedError(
            "Constraint function must be callable with model, X, period and columns"
        )


class BaseParametrizationTransformation(BaseObject):
    """
    Decision variable transformation class.

    Decision variable transforms change the parametrization of the decision
    variable. The default parametrization is a flatten array with the inputs
    for all the columns. The transform is used to change the initial guess
    passed to scipy.minimize and to inverse_transform this guess to the original
    space so that the constraints and objective function can be evaluated.

    """

    def fit(self, X, horizon, columns):
        """Fit the decision variable to the data"""
        self.horizon_ = horizon
        self.columns_ = columns
        return self._fit(X, horizon, columns)

    def transform(self, x):
        """Transform the decision variable to the original space"""
        return self._transform(x)

    def inverse_transform(self, xt):
        """Return the initial guess for the decision variable"""
        return self._inverse_transform(xt)

    def _fit(self, X: pd.DataFrame, horizon: pd.Index, columns: pd.Index):
        """Default private fit"""
        pass

    def _transform(self, x: jnp.ndarray):
        """Transform the decision variable to the original space"""
        raise NotImplementedError("Decision variable must implement transform method")

    def _inverse_transform(self, xt: jnp.ndarray):
        """Return the initial guess for the decision variable"""
        raise NotImplementedError(
            "Decision variable must implement initial_guess method"
        )


class BaseBudgetOptimizer(BaseObject):
    """Base class for budget optimization.

    Budget optimization is an optimization of a set of input variables.
    Optimizers are meant to be used in conjunction with a model. The arguments
    should be divided as follows:

    * init arguments:
        * utility function : an specification of the objective function to be
            optimized
        * contraints: a list of constraints, such as Budget, ChannelCap, etc.
        * optimization related arguments
    * optimize arguments:
        * model
        * X : pd.DataFrame
        * period : pd.PeriodIndex
        * columns : list[str]
    """

    _tags = {"backend": None}

    def __init__(
        self, constraints: list[BaseConstraint], objective: BaseOptimizationObjective
    ):
        self.constraints = constraints
        self.objective = objective
        super().__init__()

    def optimize(self, model, X: pd.DataFrame, horizon: pd.Index, columns: pd.Index):
        return self._optimize(
            model=model,
            X=X,
            horizon=horizon,
            columns=columns,
        )



"""Composition of effects (Effects that wrap other effects)."""

from typing import Any, Dict

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import pandas as pd

from prophetverse.distributions import GammaReparametrized
from prophetverse.utils.frame_to_array import series_to_tensor_or_array
from prophetverse.utils.numpyro import CacheMessenger
from .base import BaseEffect

__all__ = ["LiftExperimentLikelihood"]


class LiftExperimentLikelihood(BaseEffect):
    """Wrap an effect and applies a normal likelihood to its output.

    This class uses an input as a reference for the effect, and applies a normal
    likelihood to the output of the effect.

    Parameters
    ----------
    effect : BaseEffect
        The effect to wrap.
    lift_test_results : pd.DataFrame
        A dataframe with the lift test results. Should be in sktime format, and must
        have the same index as the input data.
    prior_scale : float
        The scale of the prior distribution for the likelihood.
    """

    _tags = {"requires_X": True, "hierarchical_prophet_compliant": False}

    def __init__(
        self,
        effect: BaseEffect,
        lift_test_results: pd.DataFrame,
        prior_scale: float,
        likelihood_scale: float = 1,
    ):

        self.effect = effect
        self.lift_test_results = lift_test_results
        self.prior_scale = prior_scale
        self.likelihood_scale = likelihood_scale

        super().__init__()

        assert self.prior_scale > 0, "prior_scale must be greater than 0"

        mandatory_columns = ["x_start", "x_end", "lift"]
        assert all(
            column in self.lift_test_results.columns for column in mandatory_columns
        ), f"lift_test_results must have the following columns: {mandatory_columns}"

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1):
        """Initialize the effect.

        This method is called during `fit()` of the forecasting model.
        It receives the Exogenous variables DataFrame and should be used to initialize
        any necessary parameters or data structures, such as detecting the columns that
        match the regex pattern.


        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.

        scale : float, optional
            The scale of the timeseries. For multivariate timeseries, this is
            a dataframe. For univariate, it is a simple float.

        Returns
        -------
        None
        """
        self.effect_ = self.effect.clone()
        self.effect_.fit(X=X, y=y, scale=scale)
        self.timeseries_scale = scale
        # super().fit(X=X, y=y, scale=scale)

    def _transform(self, X: pd.DataFrame, fh: pd.Index) -> Dict[str, Any]:
        """Prepare input data to be passed to numpyro model.

        Returns a dictionary with the data for the lift and for the inner effect.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables for the training
            time indexes, if passed during fit, or for the forecasting time indexes, if
            passed during predict.

        fh : pd.Index
            The forecasting horizon as a pandas Index.

        Returns
        -------
        Dict[str, Any]
            Dictionary with data for the lift and for the inner effect
        """
        data_dict = {}
        data_dict["inner_effect_data"] = self.effect_.transform(X, fh=fh)

        # Check if fh and self.lift_test_results have same index type
        if not isinstance(fh, self.lift_test_results.index.__class__):
            raise TypeError(
                "fh and self.lift_test_results must have the same index type"
            )

        X_lift = self.lift_test_results.reindex(fh, fill_value=jnp.nan)
        data_dict["observed_lift"] = series_to_tensor_or_array(X_lift["lift"].dropna())
        data_dict["obs_mask"] = ~jnp.isnan(series_to_tensor_or_array(X_lift["lift"]))

        # fill x_start column nan values with values from X
        X_lift["x_start"] = X_lift["x_start"].fillna(X.iloc[:, 0])
        X_lift["x_end"] = X_lift["x_end"].fillna(X.iloc[:, 0])

        data_dict["x_start"] = series_to_tensor_or_array(X_lift["x_start"])
        data_dict["x_end"] = series_to_tensor_or_array(X_lift["x_end"])

        # Get indexes of self.lift_test_results.index in X_lift
        data_dict["experiment_date_idx"] = jnp.array(
            X_lift.index.get_indexer(self.lift_test_results.index), dtype=jnp.int32
        )
        return data_dict

    def _predict(
        self, data: Dict, predicted_effects: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries.
        """
        observed_lift = data["observed_lift"].reshape((-1, 1))
        x_start = data["x_start"].reshape((-1, 1))
        x_end = data["x_end"].reshape((-1, 1))
        obs_mask = data["obs_mask"]

        with CacheMessenger():
            # Call the effect a first time
            x = self.effect_.predict(
                data=data["inner_effect_data"],
                predicted_effects=predicted_effects,
            )

            x_start_data = self.effect_._update_data(data["inner_effect_data"], x_start)
            # Get the start and end values
            y_start = self.effect_.predict(
                data=x_start_data,
                predicted_effects=predicted_effects,
            )

            x_end_data = self.effect_._update_data(data["inner_effect_data"], x_end)
            y_end = self.effect_.predict(
                data=x_end_data, predicted_effects=predicted_effects
            )

            y_end = y_end[obs_mask]
            y_start = y_start[obs_mask]

        # Calculate the delta_y
        delta_y = y_end / y_start

        with numpyro.handlers.scale(scale=self.likelihood_scale):
            distribution = GammaReparametrized(delta_y, self.prior_scale)

            # Add :ignore so that the model removes this
            # sample when organizing the output dataframe
            numpyro.sample(
                "lift_experiment:ignore",
                distribution,
                obs=observed_lift,
            )

        return x

    def _update_data(self, data, arr):
        data["inner_effect_data"] = self.effect_._update_data(
            data["inner_effect_data"], arr
        )
        return data

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        from prophetverse.effects.linear import LinearEffect

        return [
            {
                "effect": LinearEffect(),
                "lift_test_results": pd.DataFrame(
                    {
                        "x_start": [1, 2, 3],
                        "x_end": [4, 5, 6],
                        "lift": [0.1, 0.2, 0.3],
                    },
                    index=pd.date_range(start="2000-01-02", periods=3, freq="D"),
                ),
                "prior_scale": 0.1,
            }
        ]



"""Definition of Log Effect class."""

from typing import Dict, Optional

import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist
from numpyro.distributions import Distribution

from prophetverse.effects.base import (
    EFFECT_APPLICATION_TYPE,
    BaseAdditiveOrMultiplicativeEffect,
)

__all__ = ["LogEffect"]


class LogEffect(BaseAdditiveOrMultiplicativeEffect):
    """Represents a log effect as effect = scale * log(rate * data + 1).

    Parameters
    ----------
    scale_prior : Optional[Distribution], optional
        The prior distribution for the scale parameter., by default Gamma
    rate_prior : Optional[Distribution], optional
        The prior distribution for the rate parameter., by default Gamma
    effect_mode : effects_application, optional
        Either "additive" or "multiplicative", by default "multiplicative"
    """

    def __init__(
        self,
        effect_mode: EFFECT_APPLICATION_TYPE = "multiplicative",
        scale_prior: Optional[Distribution] = None,
        rate_prior: Optional[Distribution] = None,
    ):
        self.scale_prior = scale_prior
        self.rate_prior = rate_prior
        super().__init__(effect_mode=effect_mode)

        self._scale_prior = (
            self.scale_prior if scale_prior is not None else dist.Gamma(1, 1)
        )
        self._rate_prior = (
            self.rate_prior if rate_prior is not None else dist.Gamma(1, 1)
        )

    def _predict(  # type: ignore[override]
        self,
        data: jnp.ndarray,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """
        scale = numpyro.sample("log_scale", self._scale_prior)
        rate = numpyro.sample("log_rate", self._rate_prior)

        effect = scale * jnp.log(jnp.clip(rate * data + 1, 1e-8, None))

        return effect



"""Module for the inference engines."""

from .base import BaseInferenceEngine
from .map import MAPInferenceEngine, MAPInferenceEngineError
from .mcmc import MCMCInferenceEngine
from .prior import PriorPredictiveInferenceEngine
from .vi import VIInferenceEngine, VIInferenceEngineError

__all__ = [
    "BaseInferenceEngine",
    "MAPInferenceEngine",
    "MAPInferenceEngineError",
    "MCMCInferenceEngine",
    "PriorPredictiveInferenceEngine",
    "VIInferenceEngine",
    "VIInferenceEngineError",
]



"""Dataset for branding example"""

import numpy as np
from prophetverse.datasets._mmm.dataset1 import get_index, get_X
import numpyro.distributions as dist
import pandas as pd
from pathlib import Path
from prophetverse.effects import (
    HillEffect,
    LinearEffect,
    LinearFourierSeasonality,
    ChainedEffects,
    GeometricAdstockEffect,
    WeibullAdstockEffect,
    SumEffects,
    Forward,
    MultiplyEffects,
    Constant,
)
from prophetverse.effects.trend import PiecewiseLinearTrend
from prophetverse.engine.prior import PriorPredictiveInferenceEngine
from prophetverse.experimental.simulate import simulate
from prophetverse.sktime import Prophetverse
from prophetverse.utils.regex import exact, no_input_columns
from prophetverse.effects.target.univariate import NegativeBinomialTargetLikelihood
import jax.numpy as jnp
import json


def get_site_values():

    posterior_samples_path = Path(__file__).parent / Path(
        "dataset2_branding_posterior_samples.json"
    )
    with open(posterior_samples_path, "r") as f:
        posterior_samples = json.load(f)

    for key in posterior_samples.keys():
        posterior_samples[key] = jnp.array(posterior_samples[key])

    return posterior_samples


def get_groundtruth_model():
    """
    Define and configure a Prophetverse model with custom components.

    Returns
    -------
    Prophetverse
        Configured Prophetverse model.
    """

    site_samples = get_site_values()
    model = Prophetverse(
        trend=PiecewiseLinearTrend(
            changepoint_interval=100,
            changepoint_range=-100,
            remove_seasonality_before_suggesting_initial_vals=False,
        ),
        exogenous_effects=[
            (
                "yearly_seasonality",
                LinearFourierSeasonality(
                    freq="D",
                    sp_list=[365.25],
                    fourier_terms_list=[3],
                    prior_scale=0.05,
                    effect_mode="multiplicative",
                ),
                no_input_columns,
            ),
            (
                "weekly_seasonality",
                LinearFourierSeasonality(
                    freq="D",
                    sp_list=[7],
                    fourier_terms_list=[3],
                    prior_scale=0.01,
                    effect_mode="multiplicative",
                ),
                no_input_columns,
            ),
            (
                "monthly_seasonality",
                LinearFourierSeasonality(
                    freq="D",
                    sp_list=[28],
                    fourier_terms_list=[5],
                    prior_scale=0.05,
                    effect_mode="multiplicative",
                ),
                no_input_columns,
            ),
            (
                "latent/awareness",
                HillEffect(effect_mode="additive"),
                exact("ad_spend_awareness"),
            ),
            (
                "latent/awareness_adstock",
                ChainedEffects(
                    steps=[
                        ("latent_awareness", Forward("latent/awareness")),
                        ("adstock", WeibullAdstockEffect(max_lag=90)),
                    ]
                ),
                None,
            ),
            (
                "latent/baseline",
                SumEffects(
                    effects=[
                        ("trend", Forward("trend")),
                        ("yearly_seasonality", Forward("yearly_seasonality")),
                        ("weekly_seasonality", Forward("weekly_seasonality")),
                        ("monthly_seasonality", Forward("monthly_seasonality")),
                        ("latent_awareness", Forward("latent/awareness_adstock")),
                    ]
                ),
                None,
            ),
            (
                "awareness_to_sales",
                MultiplyEffects(
                    [
                        ("linear", Constant()),
                        ("awareness", Forward("latent/awareness_adstock")),
                    ]
                ),
                None,
            ),
            (
                "last_click_spend",
                MultiplyEffects(
                    [
                        (
                            "hill",
                            ChainedEffects(
                                steps=[
                                    ("adstock", GeometricAdstockEffect()),
                                    ("saturation", HillEffect(effect_mode="additive")),
                                ]
                            ),
                        ),
                        ("baseline", Forward("latent/baseline")),
                    ]
                ),
                exact("last_click_spend"),
            ),
        ],
        inference_engine=PriorPredictiveInferenceEngine(
            num_samples=1, substitute=site_samples
        ),
        scale=1,
    )
    return model


def _get_X(index, rng):
    """
    Create a DataFrame of two simulated investments with daily data.

    Parameters
    ----------
    index : pd.PeriodIndex
        The time index for the DataFrame.

    Returns
    -------
    pd.DataFrame
        DataFrame containing normalized simulated investments.
    """

    X = get_X(index, rng, power=1)
    X = X.rename(
        columns={
            "ad_spend_search": "ad_spend_awareness",
            "ad_spend_social_media": "last_click_spend",
        }
    )
    # Select 20 days to add dummy campaigns
    return X


def get_dataset(return_y_and_X_only=False):
    """
    Generate a complete dataset.

    Includes time index, exogenous data, model,
    simulated samples, observed sales, true effects, and lift test results.

    We simulate a usecase where we have two channels:
    1. A last click channel (e.g. social media) that has direct impact
         on sales and is correlated with the latent awareness channel.
    2. A latent channel (e.g. search ads) that has no direct impact on sales
         but drives awareness and therefore has an indirect impact on sales.

    Returns
    -------
    tuple
        Contains observed sales, exogenous data, lift test results, true effects,
        and the Prophetverse model.
    """

    rng = np.random.default_rng(0)
    index = get_index()
    X = _get_X(index, rng=rng)

    # We will fit, predict, and then use the prediction to create the
    # investment in last click to simulate correlation
    model = get_groundtruth_model()

    model.fit(X=X, y=pd.Series(np.zeros(X.shape[0]), index=X.index))
    y = model.predict(X=X, fh=index)

    X["last_click_spend"] = X["last_click_spend"].mean() / y.mean() * y

    model.fit(X=X, y=pd.Series(np.zeros(X.shape[0]), index=X.index))
    y = model.predict(X=X, fh=index)

    true_effect = model.predict_components(X=X, fh=index)

    X, y = X.iloc[90:], y.iloc[90:]
    true_effect = true_effect.iloc[90:]
    if return_y_and_X_only:
        return y, X
    return y, X, true_effect, model



"""Optimizers module."""

from .optimizer import (
    AdamOptimizer,
    BaseOptimizer,
    CosineScheduleAdamOptimizer,
    LBFGSSolver,
    _NumPyroOptim,
)

__all__ = [
    "AdamOptimizer",
    "BaseOptimizer",
    "CosineScheduleAdamOptimizer",
    "_LegacyNumpyroOptimizer",
    "_NumPyroOptim",
    "_OptimizerFromCallable",
    "LBFGSSolver",
]



"""Module that stores abstract class of effects."""

from typing import Any, Dict, Literal, Optional
import numpyro
import jax.numpy as jnp
import pandas as pd
from skbase.base import BaseObject
import numpyro
from prophetverse.utils.deprecation import deprecation_warning
from prophetverse.utils import series_to_tensor_or_array
from prophetverse.utils.numpyro import CacheMessenger
from collections import OrderedDict, defaultdict
import copy

__all__ = ["BaseEffect", "BaseAdditiveOrMultiplicativeEffect"]


EFFECT_APPLICATION_TYPE = Literal["additive", "multiplicative"]


class BaseEffect(BaseObject):
    """Base class for effects.

    Effects are objects which are responsible for preparing the data and applying
    a specific effect to the forecast. During preparation of the data (which happens in
    `transform` method), the effect receives the exogenous variables dataframe and can
    use them to prepare the jax arrays that will be used at inference time. During
    inference time, the `predict` method is called, and it should output a new component
    to the additive model of Prophetverse.

    Remember that Prophetverse's models are Generalized Additive Models, which are
    composed of many terms summed together to form the final forecast. Each term is
    represented by an effect.

    Children classes should implement the following methods:


    * `_fit` (optional): This method is called during fit() of the forecasting  and
    should be used to initialize any necessary parameters or data structures.
    It receives the exogenous variables dataframe X, the series `y`, and the scale
    factor `scale` that was used to scale the timeseries.

    * `_transform` (optional): This method receives the exogenous variables
    dataframe, and should return an object containing the data needed for the
    effect. This object will be passed to the predict method as `data`. By default
    the columns of the dataframe that match the regex pattern are selected, and the
    result is converted to a `jnp.ndarray`.

    * `_predict` (mandatory): This method receives the output of `_transform` and
    all previously computed effects. It should return the effect values as a
    `jnp.ndarray`


    Parameters
    ----------
    id : str, optional
        The id of the effect, by default "". Used to identify the effect in the model.
    regex : Optional[str], optional
        A regex pattern to match the columns of the exogenous variables dataframe,
        by default None. If None, and _tags["requires_X"] is True, the
        effect will be skipped if no columns are found.
    effect_mode : EFFECT_APPLICATION_TYPE, optional
        The mode of the effect, either "additive" or "multiplicative", by default
        "multiplicative". If "multiplicative", the effect multiplies the trend values
        before returning them.


    Attributes
    ----------
    should_skip_predict : bool
        If True, the effect should be skipped during prediction. This is determined by
        the `requires_X` tag and the presence of input feature columns
        names. If the tag is True and there are no input feature columns names, the
        effect should be skipped during prediction.
    """

    _tags = {
        # Can handle panel data?
        "capability:panel": False,
        # Can be used with hierarchical Prophet?
        "capability:panel": False,
        # Can handle multiple input feature columns?
        "capability:multivariate_input": False,
        # If no columns are found, should
        # _predict be skipped?
        "requires_X": True,
        # Should only the indexes related to the forecasting horizon be passed to
        # _transform?
        "filter_indexes_with_forecating_horizon_at_transform": True,
        # Is fit() required before calling transform()?
        "requires_fit_before_transform": False,
        # should this effect be applied to `y` (target) or
        # `X` (exogenous variables)?
        "applies_to": "X",
        # Does this effect implement hyperpriors across panel?
        "feature:panel_hyperpriors": False,
    }

    def __init__(self):
        self._is_fitted: bool = False
        super().__init__()
        self._broadcasted = None

    def fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1.0):
        """Initialize the effect.

        This method is called during `fit()` of the forecasting model.
        It receives the Exogenous variables DataFrame and should be used to initialize
        any necessary parameters or data structures, such as detecting the columns that
        match the regex pattern.

        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.

        scale : float, optional
            The scale of the timeseries. For multivariate timeseries, this is
            a dataframe. For univariate, it is a simple float.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the effect does not support multivariate data and the DataFrame has more
            than one level of index.
        """

        self.columns_ = None
        if X is not None:
            self.columns_ = X.columns.tolist()

        data = X if self.get_tag("applies_to", "X") == "X" else y

        if (
            data is not None
            and len(data.columns) > 1
            and not self.get_tag("capability:multivariate_input", False)
        ):
            self._set_broadcasting_attributes(data)
            self._broadcast("fit", X=X, y=y, scale=scale)

        elif (
            data is not None
            and data.index.nlevels > 1
            and not self.get_tag("capability:panel", False)
        ):
            self._set_panel_broadcasting_attributes(data)
            self._broadcast("fit", X=X, y=y, scale=scale)
        else:
            self._fit(y=y, X=X, scale=scale)
        self._is_fitted = True

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1.0):
        """Customize the initialization of the effect.

        This method is called by the `fit()` method and can be overridden by
        subclasses to provide additional initialization logic.

        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.

        scale : float, optional
            The scale of the timeseries. For multivariate timeseries, this is
            a dataframe. For univariate, it is a simple float.
        """
        pass

    def transform(
        self,
        X: pd.DataFrame,
        fh: pd.Index,
    ) -> Any:
        """Prepare input data to be passed to numpyro model.

        This method receives the Exogenous variables DataFrame and should return a
        the data needed for the effect. Those data will be passed to the `predict`
        method as `data` argument.

        The private methods `_transform` should always return one of the
        following:

        * a `jnp.ndarray`, with an array or tensor with the data
          from the exogenous variables
        * a tuple, where the first element is a `jnp.ndarray` with the data
            from the exogenous variables, and the rest of the elements are
        * a dict, where one of the keys must be "data" and the value

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables for the training
            time indexes, if passed during fit, or for the forecasting time indexes, if
            passed during predict.

        fh : pd.Index
            The forecasting horizon as a pandas Index.

        Returns
        -------
        Any
            Any object containing the data needed for the effect. The object will be
            passed to `predict` method as `data` argument.

        Raises
        ------
        ValueError
            If the effect has not been fitted.
        """
        if not self._is_fitted and self.get_tag("requires_fit_before_transform", True):
            raise ValueError("You must call fit() before calling this method")

        if self.get_tag("filter_indexes_with_forecating_horizon_at_transform", True):
            # Filter when index level -1 is in fh
            if X is not None:
                X = X.loc[X.index.get_level_values(-1).isin(fh)]

        if not self._is_fitted and X is not None:
            if len(X.columns) > 1 and not self.get_tag(
                "capability:multivariate_input", False
            ):
                if not self._is_fitted:
                    # Since the broadcasting attributes are set during fit,
                    # we need to set them
                    self._set_broadcasting_attributes(X)

            elif X.index.nlevels > 1 and not self.get_tag("capability:panel", False):

                # Since the broadcasting attributes are set during fit,
                # we need to set them
                self._set_panel_broadcasting_attributes(X)

        if self._broadcasted is not None:
            return self._broadcast("transform", X=X, fh=fh)

        return self._transform(X, fh)

    def _broadcast(self, methodname: str, X, **kwargs):
        """
        Broadcasts a method to  handle multiple columns of the input DataFrame.

        Parameters
        ----------
        methodname : str
            The name of the method to be called.
        *args : tuple
            Positional arguments to be passed to the method.
        **kwargs : dict
            Keyword arguments to be passed to the method.

        Returns
        -------
        jnp.ndarray
        """
        if self._broadcasted == "panel":
            return self._broadcast_panel(X, methodname, **kwargs)
        if self._broadcasted == "columns":
            return self._broadcast_columns(X, methodname, **kwargs)
        raise ValueError(
            f"Broadcasting not set for {self.__class__.__name__}. "
            "Please call fit() before calling transform()."
        )

    def _broadcast_columns(self, X: pd.DataFrame, methodname: str, **kwargs):
        """
        Broadcasts a method to handle multiple columns of the input DataFrame.

        This method iterates over the columns of the DataFrame and calls the
        specified method on each column, collecting the results in a list.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables.
        methodname : str
            The name of the method to be called on each column.
        **kwargs : dict
            Additional keyword arguments to be passed to the method.
        Returns
        -------
        list
            A list of outputs from the method applied to each column.
        """
        outputs = []
        for column in self.effects_.keys():
            X_ = X[[column]]
            effect_ = self.effects_[column]
            xt = getattr(effect_, methodname)(X=X_, **kwargs)
            outputs.append(xt)
        return outputs

    def _broadcast_panel(self, X: pd.DataFrame, methodname: str, **kwargs):
        """
        Broadcasts a method to handle panel data of the input DataFrame.

        This method iterates over the idx of the DataFrame and calls the
        specified method on each idx, collecting the results in a list.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables.
        methodname : str
            The name of the method to be called on each column.
        **kwargs : dict
            Additional keyword arguments to be passed to the method.
        Returns
        -------
        list
            A list of outputs from the method applied to each column.
        """
        outputs = []

        for i, idx in enumerate(self.effects_.keys()):

            X_ = None
            if X is not None:
                X_ = X.loc[idx]
            new_kwargs = kwargs.copy()
            if "y" in kwargs:
                new_kwargs["y"] = kwargs["y"].loc[idx]
            if "scale" in kwargs:
                scale_value = kwargs["scale"]
                is_pd_obj = isinstance(scale_value, (pd.Series, pd.DataFrame))
                if is_pd_obj:
                    scale_value = scale_value.loc[idx]
                if isinstance(scale_value, (pd.Series, pd.DataFrame)) and scale_value.size == 1:
                    scale_value = scale_value.values.item()
                new_kwargs["scale"] = scale_value
            effect_ = self.effects_[idx]
            xt = getattr(effect_, methodname)(X=X_, **new_kwargs)
            outputs.append(xt)
        return outputs

    def _transform(
        self,
        X: pd.DataFrame,
        fh: pd.Index,
    ) -> Any:
        """Prepare input data to be passed to numpyro model.

        This method receives the Exogenous variables DataFrame and should return a
        the data needed for the effect. Those data will be passed to the `predict`
        method as `data` argument.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables for the training
            time indexes, if passed during fit, or for the forecasting time indexes, if
            passed during predict.

        fh : pd.Index
            The forecasting horizon as a pandas Index.

        Returns
        -------
        Any
            Any object containing the data needed for the effect. The object will be
            passed to `predict` method as `data` argument.
        """
        array = series_to_tensor_or_array(X)
        return array

    def predict(
        self,
        data: Dict,
        predicted_effects: Optional[Dict[str, jnp.ndarray]] = None,
        params: Optional[Dict[str, jnp.ndarray]] = None,
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """
        if predicted_effects is None:
            predicted_effects = {}

        if isinstance(data, list):
            if self._broadcasted == "columns":
                x = 0
                for i, _data in enumerate(data):
                    effect_ = self.effects_[self.columns_[i]]
                    with numpyro.handlers.scope(prefix=str(self.columns_[i])):
                        out = effect_.predict(
                            data=_data,
                            predicted_effects=predicted_effects,
                            params=params,
                        )
                    out = numpyro.deterministic(str(self.columns_[i]), out)
                    x += out
            elif self._broadcasted == "panel":
                x = []
                for i, _data in enumerate(data):
                    effect_ = self.effects_[self.idxs_[i]]
                    idx = self.idxs_[i]
                    if isinstance(idx, (tuple, list)):
                        idx = ",".join(idx)
                    else:
                        idx = str(idx)
                    prefix = f"panel-{i}"
                    _predicted_effects = {k: v[i] for k, v in predicted_effects.items()}
                    with numpyro.handlers.scope(prefix=prefix):
                        out = effect_.predict(
                            data=_data,
                            predicted_effects=_predicted_effects,
                            params=params,
                        )
                    x.append(out)

                # If output is None, then the effect
                # shouldn't output anything.
                if x[0] is None:
                    return None
                # If out is of shape (N, T), x will be
                # (D, N, T)
                x = jnp.stack(x, axis=0)
                x = numpyro.deterministic("panel_effects", x)
        else:
            x = self._predict(data, predicted_effects, params)
        return x

    def _predict(
        self, data: Dict, predicted_effects: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """
        raise NotImplementedError("Subclasses must implement _predict()")

    def __call__(
        self,
        data: Dict,
        predicted_effects: Dict[str, jnp.ndarray],
        params: Optional[Dict[str, jnp.ndarray]] = None,
    ) -> jnp.ndarray:
        """Run the processes to calculate effect as a function."""
        return self.predict(data=data, predicted_effects=predicted_effects)

    def _update_data(self, data: jnp.ndarray, arr: jnp.ndarray):
        """
        Update the data obtained from .transform with a new array.

        This method is used during optimization to update the transformed
        data, given an array with new information.

        If
        """

        if isinstance(data, jnp.ndarray):
            return arr
        if isinstance(data, tuple):
            return (arr, *data[1:])
        if isinstance(data, dict):
            data = data.copy()
            data["data"] = arr
            return data
        if isinstance(data, list) and self._broadcasted == "columns":
            out = []
            for i, d in enumerate(data):
                out.append(self._update_data(d, arr[:, i].reshape((-1, 1))))
            return out
        if isinstance(data, list) and self._broadcasted == "panel":
            out = []
            for i, d in enumerate(data):
                out.append(self._update_data(d, arr[i]))
            return out
        raise ValueError(
            f"Unexpected data type {type(data)}. "
            "Expected jnp.ndarray, tuple, dict or list."
        )

    def _set_broadcasting_attributes(self, X):
        """
        Set broadcasting attributes for the effect.

        This method is called during the `fit` method to set the
        broadcasting attributes for the effect, or during `transform`
        of the method does not require fitting before transform.
        """
        self.effects_ = OrderedDict((column, self.clone()) for column in X.columns)
        self.columns_ = X.columns.tolist()
        self._broadcasted = "columns"

    def _set_panel_broadcasting_attributes(self, X):

        if X.index.nlevels == 1:
            return
        elif self.get_tag("capability:panel", False):
            return
        # Else proceed to broadcasting by panel

        if self._broadcasted == "columns":
            raise ValueError(
                "Cannot set panel broadcasting attributes when "
                + "broadcasting by columns is already set."
            )

        idxs = X.index.droplevel(-1).unique().tolist()
        self.effects_ = OrderedDict((idx, self.clone()) for idx in idxs)
        self.idxs_ = idxs
        self._broadcasted = "panel"

    def _get_distribution_params(self, params):
        """
        Get numpyro distribution parameters.

        Parameters
        ----------
        params : dict
            A dictionary containing the parameters of the effect.

        Returns
        -------
        dict
            A dictionary containing the distribution parameters of the effect.
        """
        distribution_params = {}
        for param_name, param_value in params.items():
            if isinstance(param_value, numpyro.distributions.Distribution):
                # Get init args of param_value
                init_args = param_value.__init__.__code__.co_varnames[1:]
                for argname in init_args:
                    try:
                        distribution_params[param_name + "__" + argname] = getattr(
                            param_value, argname
                        )
                    except AttributeError:
                        pass
        return distribution_params

    def get_params(self, deep=True):
        """
        Override get_params to include distribution parameters.

        Parameters
        ----------
        deep : bool, optional
            If True, will return the parameters of the effect and its sub-
            estimators,
            by default True.

        Returns
        -------
        dict
            A dictionary containing the parameters of the effect, including
            distribution parameters.
        """
        params = super().get_params(deep=deep)
        if not deep:
            return params
        new_params = self._get_distribution_params(params)
        return {**params, **new_params}

    def set_params(self, **params):
        """
        Extend set_params to handle distribution parameters.
        """

        all_params = self.get_params(deep=False)
        distribution_params = self._get_distribution_params(all_params)

        # We list all distribution parameters and their args
        distribution_params_to_update = defaultdict(dict)
        for param_name, param_value in params.items():
            if param_name in distribution_params:
                base_param_name, distribution_argname = param_name.split("__")
                distribution_params_to_update[base_param_name] = {
                    **distribution_params_to_update[base_param_name],
                    distribution_argname: param_value,
                }

        # We update the distribution parameters with the new values
        # and set them to the effect
        for param_to_update, distribution_args in distribution_params_to_update.items():
            # Create the distribution object
            distribution_obj = copy.deepcopy(getattr(self, param_to_update))
            for argname, value in distribution_args.items():
                setattr(distribution_obj, argname, value)
            # Update the distribution param
            self.set_params(**{param_to_update: distribution_obj})

        not_distribution_params = {
            k: v for k, v in params.items() if k not in distribution_params
        }
        return super().set_params(**not_distribution_params)

    def sample_prior(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame = None,
        num_samples=10,
        predicted_effects=None,
        seed=42,
        as_pandas=False,
    ):
        """
        Sample from the prior distribution.

        Parameters
        ----------
        y: pd.DataFrame
            The target variable.
        X : pd.DataFrame
            The input data for which to sample from the prior distribution.
        num_samples : int, optional
            The number of samples to draw from the prior distribution, by default 10.
        predicted_effects : Optional[Dict[str, jnp.ndarray]], optional
            A dictionary of predicted effects to condition on, by default None.
        seed : int, optional
            The random seed to use for sampling, by default 42.
        as_pandas: bool, optional
            Whether to return the samples as a pandas DataFrame, by default False.

        Returns
        -------
        jnp.ndarray
            The output of shape (X.shape[0], num_samples)
        """

        from prophetverse.engine.prior import PriorPredictiveInferenceEngine

        # Clone and fit
        instance = self.clone()
        instance.fit(y=y, X=X)

        z = X
        if instance.get_tag("applies_to", "X") == "y":
            z = y
        predict_data = instance.transform(z, fh=X.index.get_level_values(-1).unique())

        def wrapper(**kwargs):
            with CacheMessenger(), numpyro.handlers.scope(prefix="effect"):
                out = instance.predict(**kwargs)
            return numpyro.deterministic("__out", out)

        # Get prior engine
        engine = PriorPredictiveInferenceEngine(
            return_sites="all", num_samples=num_samples
        )
        engine.infer(wrapper, data=predict_data, predicted_effects=predicted_effects)
        samples = engine.predict(data=predict_data, predicted_effects=predicted_effects)
        samples = samples["__out"]
        if not as_pandas:
            return samples
        return pd.concat(
            {
                i: pd.DataFrame(samples[i].flatten(), index=X.index)
                for i in range(samples.shape[0])
            }
        )


class BaseAdditiveOrMultiplicativeEffect(BaseEffect):
    """
    Base class for effects that can be applied in additive or multiplicative mode.

    In additive mode, the effect is directly returned. In multiplicative mode,
    the effect is multiplied by the trend before being returned. In other words:

    Additive effect: effect = _predict(trend, **kwargs)
    Multiplicative effect: effect = trend * _predict(trend, **kwargs)

    Parameters
    ----------
    id : str, optional
        The id of the effect, by default "".
    regex : Optional[str], optional
        A regex pattern to match the columns of the exogenous variables dataframe,
        by default None. If None, and _tags["requires_X"] is True, the
        effect will be skipped if no columns are found.
    effect_mode : EFFECT_APPLICATION_TYPE, optional. Can be "additive"
        or "multiplicative".
    """

    def __init__(self, effect_mode="additive", base_effect_name: str = "trend"):

        self.effect_mode = effect_mode
        self.base_effect_name = base_effect_name

        if effect_mode not in ["additive", "multiplicative"]:
            raise ValueError(
                f"Invalid effect mode: {effect_mode}. "
                + "Effect mode must be 'additive' or 'multiplicative'."
            )

        super().__init__()

        if not isinstance(self.base_effect_name, list):
            self._base_effect_name = [self.base_effect_name]
        else:
            self._base_effect_name = self.base_effect_name

    def predict(
        self,
        data: Any,
        predicted_effects: Optional[Dict[str, jnp.ndarray]] = None,
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """

        x = super().predict(
            data=data, predicted_effects=predicted_effects, *args, **kwargs
        )

        if self.effect_mode == "additive":
            return x

        base_effect = 0
        for base_effect_name in self._base_effect_name:
            if base_effect_name not in predicted_effects:
                raise ValueError(
                    f"BaseAdditiveOrMultiplicativeEffect requires {base_effect_name} in"
                    + " predicted_effects"
                )

            base_effect += predicted_effects[base_effect_name]

        if base_effect.ndim == 1:
            base_effect = base_effect.reshape((-1, 1))
        x = x.reshape(base_effect.shape)
        return base_effect * x



"""Utilities to handle deprecation."""

import warnings


def deprecation_warning(obj_name, current_version, extra_message=""):
    """
    Generate a deprecation warning for an object.

    Parameters
    ----------
        obj_name (str): The name of the object to be deprecated.
        current_version (str): The current version in the format 'major.minor.patch'.

    Returns
    -------
        str: A deprecation warning message.
    """
    try:
        # Parse the current version into components
        major, minor, patch = map(int, current_version.split("."))
        # Calculate the deprecation version (1 minor releases ahead)
        new_minor = minor + 1
        deprecation_version = f"{major}.{new_minor}.0"
        # Return the deprecation warning
        warnings.warn(
            f"Warning: '{obj_name}' is deprecated and will be removed in version"
            f" {deprecation_version}. Please update your code to avoid issues. "
            f"{extra_message}",
            FutureWarning,
            stacklevel=2,
        )
    except ValueError:
        raise ValueError("Invalid version format. Expected 'major.minor.patch'.")



from prophetverse.budget_optimization.base import (
    BaseOptimizationObjective,
)
import jax.numpy as jnp

__all__ = [
    "MaximizeROI",
    "MaximizeKPI",
    "MinimizeBudget",
]


class MaximizeROI(BaseOptimizationObjective):
    """
    Maximize return on investment (ROI) objective function.
    """

    _tags = {
        "name": "MaxROI",
        "backend": "scipy",
    }

    def __init__(self):
        super().__init__()

    def _objective(self, x: jnp.ndarray, budget_optimizer):
        """
        Compute objective function value from `obs` site

        Parameters
        ----------
        obs : jnp.ndarray
            Observed values

        Returns
        -------
        float
            Objective function value
        """
        obs = budget_optimizer.predictive_(x)
        obs = obs.mean(axis=0).squeeze(-1)
        obs_horizon = obs[..., budget_optimizer.horizon_idx_]
        total_return = obs_horizon.sum()
        spend = x.sum()

        return -total_return / spend


class MaximizeKPI(BaseOptimizationObjective):
    """
    Maximize the KPI objective function.
    """

    def __init__(self):
        super().__init__()

    def _objective(self, x: jnp.ndarray, budget_optimizer):
        """
        Compute objective function value from `obs` site

        Parameters
        ----------
        obs : jnp.ndarray
            Observed values

        Returns
        -------
        float
            Objective function value
        """
        obs = budget_optimizer.predictive_(x)
        obs = obs.mean(axis=0).squeeze(-1)
        obs_horizon = obs[..., budget_optimizer.horizon_idx_]
        obs_horizon = obs_horizon.sum(axis=0)

        value = -obs_horizon.sum()
        return value


class MinimizeBudget(BaseOptimizationObjective):
    """
    Minimize budget constraint objective function.
    """

    def __init__(self, scale=1):
        self.scale = scale
        super().__init__()

    def _objective(self, x: jnp.ndarray, budget_optimizer):
        """
        Compute objective function value from `obs` site

        Parameters
        ----------
        obs : jnp.ndarray
            Observed values

        Returns
        -------
        float
            Objective function value
        """
        total_investment = x.sum() / self.scale

        return total_investment



from prophetverse.effects.base import BaseEffect


class BaseTargetEffect(BaseEffect):
    """Base class for effects."""

    _tags = {
        # Supports multivariate data? Can this
        # Effect be used with Multiariate prophet?
        "hierarchical_prophet_compliant": False,
        # If no columns are found, should
        # _predict be skipped?
        "requires_X": False,
        # Should only the indexes related to the forecasting horizon be passed to
        # _transform?
        "filter_indexes_with_forecating_horizon_at_transform": True,
        # Should the effect be applied to the target variable?
        "applies_to": "y",
    }

    def _transform(self, X, fh):
        if X is not None:
            return super()._transform(X, fh)
        return None



"""Store input to predicted_effect"""

from typing import Any, Dict, Optional

import jax.numpy as jnp
import pandas as pd

from prophetverse.effects.base import BaseEffect

__all__ = ["Identity"]


class Identity(BaseEffect):
    """
    Return the input as the predicted effect.

    This effect simply returns the input data during prediction without any
    modification.
    """

    _tags = {
        "requires_X": True,  # Default value, will be overridden in __init__
        "applies_to": "X",
    }

    def __init__(self):

        super().__init__()

    def _transform(self, X, fh):
        if X is None:
            raise ValueError("Input X cannot be None in _transform method.")
        return super()._transform(X, fh)

    def _predict(
        self,
        data: Any,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """Return the data

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method (ignored).

        predicted_effects : Dict[str, jnp.ndarray]
            A dictionary containing the predicted effects (ignored).

        Returns
        -------
        jnp.ndarray
            Zero.
        """
        return data



from typing import Any, Dict, Optional

import jax.numpy as jnp
import pandas as pd

import numpyro
import numpyro.distributions as dist
from prophetverse.effects.base import BaseEffect
from prophetverse.utils.frame_to_array import series_to_tensor_or_array
from prophetverse.distributions import GammaReparametrized
from prophetverse.effects.target.base import BaseTargetEffect


class MultivariateNormal(BaseTargetEffect):
    """Base class for effects."""

    _tags = {
        # Supports multivariate data? Can this
        # Effect be used with Multiariate prophet?
        "hierarchical_prophet_compliant": True,
        "capability:panel": True,
        # If no columns are found, should
        # _predict be skipped?
        "requires_X": False,
        # Should only the indexes related to the forecasting horizon be passed to
        # _transform?
        "filter_indexes_with_forecating_horizon_at_transform": True,
    }

    def __init__(self, noise_scale=0.05, correlation_matrix_concentration=1):

        self.noise_scale = noise_scale
        self.correlation_matrix_concentration = correlation_matrix_concentration
        super().__init__()

    def _predict(
        self,
        data: Any,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        params : Dict[str, jnp.ndarray]
            A dictionary containing the sampled parameters of the effect.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """

        y = data

        mean = 0
        for effect_name, effect in predicted_effects.items():
            # Skip latent effects (effects whose names start with "latent/")
            if not effect_name.startswith("latent/"):
                mean += effect

        if y is not None:
            y = y.squeeze(-1).T

        mean = numpyro.deterministic("mean", mean)

        is_single_series = mean.shape[0] == 1 or mean.ndim < 3
        if is_single_series:

            mean = mean.reshape((-1, 1))
            std_observation = numpyro.sample(
                "std_observation", dist.HalfNormal(jnp.array(self.noise_scale))
            )

            with numpyro.plate("time", mean.shape[-1], dim=-2):
                numpyro.sample("obs", dist.Normal(mean, std_observation), obs=y)

        else:

            std_observation = numpyro.sample(
                "std_observation",
                dist.HalfNormal(jnp.array([self.noise_scale] * mean.shape[0])),
            )

            correlation_matrix = numpyro.sample(
                "corr_matrix",
                dist.LKJCholesky(
                    mean.shape[0],
                    concentration=self.correlation_matrix_concentration,
                ),
            )

            cov_mat = (
                jnp.diag(std_observation)
                @ correlation_matrix
                @ jnp.diag(std_observation)
            )

            cov_mat = jnp.tile(jnp.expand_dims(cov_mat, axis=0), (mean.shape[1], 1, 1))

            with numpyro.plate("time", mean.shape[-1], dim=-2):
                numpyro.sample(
                    "obs",
                    dist.MultivariateNormal(mean.squeeze(-1).T, scale_tril=cov_mat),
                    obs=y,
                )

        return jnp.zeros_like(mean)



"""Functions that perform algebraic operations."""

from typing import Union

import jax.numpy as jnp
from numpy.typing import NDArray


def matrix_multiplication(
    data: Union[jnp.ndarray, NDArray], coefficients: Union[jnp.ndarray, NDArray]
) -> Union[jnp.ndarray, NDArray]:
    """Perform matrix multiplication between two matrixes.

    Parameters
    ----------
    data : Union[jnp.ndarray, NDArray]
        Array to be multiplied.
    coefficients : Union[jnp.ndarray, NDArray]
        Array of coefficients used at matrix multiplication.

    Returns
    -------
    Union[jnp.ndarray, NDArray]
        Matrix multiplication between data and coefficients.
    """
    return data @ coefficients.reshape((-1, 1))


def _exponent_safe(data: Union[jnp.ndarray, NDArray], exponent: float) -> jnp.ndarray:
    """Exponentiate an array without numerical errors replacing zeros with ones.

    Parameters
    ----------
    data : Union[jnp.ndarray, NDArray]
        Array to be exponentiated.
    exponent : float
        Expoent numerical value.

    Returns
    -------
    jnp.ndarray
        Exponentiated array.
    """
    exponent_safe = jnp.where(data == 0, 1, data) ** exponent
    return jnp.where(data == 0, 0, exponent_safe)



"""Utils for inference engines."""

from typing import Dict

import jax.numpy as jnp
import numpy as np

from prophetverse.exc import ConvergenceError


def assert_mcmc_converged(summary: Dict[str, Dict[str, jnp.ndarray]], max_r_hat: float):
    """Assert that an MCMC program has converged.

    Parameters
    ----------
    summary: Dict
        MCMC trace summary.

    max_r_hat: float
        Maximum allowed r_hat.

    Returns
    -------
        Nothing.

    Raises
    ------
    ConvergenceError
    """
    for name, parameter_summary in summary.items():
        # NB: some variables have deterministic elements (s.a. samples from LKJCov).
        mask = np.isnan(parameter_summary["n_eff"])
        r_hat = parameter_summary["r_hat"][~mask]

        if (r_hat <= max_r_hat).all():
            continue

        # TODO: might be better to print entire
        #  summary instead of just which parameter didn't converge
        raise ConvergenceError(f"Parameter '{name}' did not converge! R_hat: {r_hat}")

    return



"""Contains the implementation of the HierarchicalProphet forecaster."""

from typing import Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sktime.forecasting.base import ForecastingHorizon
from sktime.transformations.base import BaseTransformer
from sktime.transformations.hierarchical.aggregate import Aggregator

from prophetverse.effects.target.multivariate import MultivariateNormal
from prophetverse.sktime.base import BaseEffect, BaseProphetForecaster
from prophetverse.utils import loc_bottom_series, reindex_time_series, series_to_tensor

from ._expand_column_per_level import ExpandColumnPerLevel


class HierarchicalProphet(BaseProphetForecaster):
    """A Bayesian hierarchical time series forecasting model based on Meta's Prophet.

    This method forecasts all bottom series in a hierarchy at once, using a
    MultivariateNormal as the likelihood function and LKJ priors for the correlation
    matrix.

    This forecaster is particularly interesting if you want to fit shared coefficients
    across series. In that case, `shared_features` parameter should be a list of
    feature names that should have that behaviour.

    Parameters
    ----------
    trend : Union[BaseEffect, str], default="linear"
        Trend component of the model.
    feature_transformer : BaseTransformer, default=None
        Transformer for features preprocessing.
    exogenous_effects : optional, default=None
        Effects to model exogenous variables.
    default_effect : optional, default=None
        Default effect specification.
    shared_features : optional, default=None
        Features shared across time series.
    noise_scale : float, default=0.05
        Scale parameter for the noise distribution.
    correlation_matrix_concentration : float, default=1.0
        Concentration parameter for the correlation matrix.
    rng_key : optional, default=None
        Random number generator key.
    inference_engine : optional, default=None
        Engine used for inference.

    Examples
    --------
    >>> from sktime.forecasting.naive import NaiveForecaster
    >>> from sktime.transformations.hierarchical.aggregate import Aggregator
    >>> from sktime.utils._testing.hierarchical import _bottom_hier_datagen
    >>> from prophetverse.sktime.multivariate import HierarchicalProphet
    >>> agg = Aggregator()
    >>> y = _bottom_hier_datagen(
    ...     no_bottom_nodes=3,
    ...     no_levels=1,
    ...     random_seed=123,
    ...     length=7,
    ... )
    >>> y = agg.fit_transform(y)
    >>> forecaster = HierarchicalProphet()
    >>> forecaster = forecaster.fit(y)
    >>> y_pred = forecaster.predict(fh=[1])
    """

    _tags = {
        # packaging info
        # --------------
        "authors": "felipeangelimvieira",
        "maintainers": "felipeangelimvieira",
        "python_dependencies": "prophetverse",
        # estimator type
        "scitype:y": "univariate",
        "y_inner_mtype": [
            "pd.DataFrame",
            "pd-multiindex",
            "pd_multiindex_hier",
        ],
        "X_inner_mtype": [
            "pd.DataFrame",
            "pd-multiindex",
            "pd_multiindex_hier",
        ],  # which types do _fit, _predict, assume for X?
        "requires-fh-in-fit": False,
        "X-y-must-have-same-index": False,
        "fit_is_empty": False,
        "capability:pred_int": True,
        "capability:pred_int:insample": True,
    }

    def __init__(
        self,
        trend: Union[BaseEffect, str] = "linear",
        feature_transformer: BaseTransformer = None,
        exogenous_effects=None,
        default_effect=None,
        shared_features=None,
        noise_scale=0.05,
        correlation_matrix_concentration=1.0,
        rng_key=None,
        inference_engine=None,
        likelihood=None,
    ):

        self.noise_scale = noise_scale
        self.shared_features = shared_features
        self.feature_transformer = feature_transformer
        self.correlation_matrix_concentration = correlation_matrix_concentration
        self.likelihood = likelihood

        super().__init__(
            # Trend
            trend=trend,
            # Exog effects
            default_effect=default_effect,
            exogenous_effects=exogenous_effects,
            # Base Bayesian forecaster
            rng_key=rng_key,
            inference_engine=inference_engine,
        )

        self._validate_hyperparams()

    def _validate_hyperparams(self):
        """Validate the hyperparameters of the HierarchicalProphet forecaster."""
        super()._validate_hyperparams()

        if self.noise_scale <= 0:
            raise ValueError("noise_scale must be greater than 0.")
        if self.correlation_matrix_concentration <= 0:
            raise ValueError("correlation_matrix_concentration must be greater than 0.")

    def _get_fit_data(self, y, X, fh):
        """
        Prepare the data for the NumPyro model.

        Parameters
        ----------
        y: pd.DataFrame
            Training target time series.
        X: pd.DataFrame
            Training exogenous variables.
        fh: ForecastingHorizon
            Forecasting horizon.

        Returns
        -------
        dict
            A dictionary containing the model data.
        """
        # Handling series without __total indexes
        self.aggregator_ = Aggregator()
        self.original_y_indexes_ = y.index
        fh = y.index.get_level_values(-1).unique()
        y = self.aggregator_.fit_transform(y)

        # Updating internal _y of sktime because BaseBayesianForecaster
        # uses it to convert
        # Forecast Horizon into multiindex correcly
        self.internal_y_indexes_ = y.index

        # Convert inputs to array, including the time index
        y_bottom = loc_bottom_series(y)
        y_bottom_arrays = series_to_tensor(y_bottom)

        # If no exogenous variables, create empty DataFrame
        # Else, aggregate exogenous variables and transform them
        if X is None or X.columns.empty:
            X = pd.DataFrame(index=y.index)

        X_bottom = loc_bottom_series(X)

        if self.feature_transformer is not None:
            X_bottom = self.feature_transformer.fit_transform(X_bottom)

        self._has_exogenous_variables = (
            X_bottom is not None and not X_bottom.columns.empty
        )

        if self._has_exogenous_variables:
            shared_features = self.shared_features
            if shared_features is None:
                shared_features = []

            self.expand_columns_transformer_ = ExpandColumnPerLevel(
                X_bottom.columns.difference(shared_features).to_list()
            ).fit(X_bottom)
            X_bottom = self.expand_columns_transformer_.transform(X_bottom)

        else:
            self._exogenous_effects_and_columns = {}
            exogenous_data = {}

        # Trend model
        self.trend_model_ = self._trend.clone()
        self.trend_model_.fit(X=X_bottom, y=y_bottom, scale=self._scale)

        if self.likelihood is not None:
            self.likelihood_model_ = self.likelihood.clone()
        else:
            self.likelihood_model_ = MultivariateNormal(
                noise_scale=self.noise_scale,
                correlation_matrix_concentration=self.correlation_matrix_concentration,
            )

        self.likelihood_model_.fit(X=X_bottom, y=y_bottom, scale=self._scale)

        trend_data = self.trend_model_.transform(X=X_bottom, fh=fh)
        target_data = self.likelihood_model_.transform(X=y_bottom, fh=fh)

        self._fit_effects(X_bottom, y_bottom)
        exogenous_data = self._transform_effects(X_bottom, fh=fh, y=y_bottom)

        self.fit_and_predict_data_ = {
            "trend_model": self.trend_model_,
            "target_model": self.likelihood_model_,
            "exogenous_effects": self.non_skipped_exogenous_effect,
        }

        return dict(
            y=y_bottom_arrays,
            data=exogenous_data,
            trend_data=trend_data,
            target_data=target_data,
            **self.fit_and_predict_data_,
        )

    def _get_predict_data(self, X: pd.DataFrame, fh: ForecastingHorizon) -> np.ndarray:
        """Generate samples for the given exogenous variables and forecasting horizon.

        Parameters
        ----------
        X: pd.DataFrame
            Exogenous variables.
        fh: ForecastingHorizon
            Forecasting horizon.

        Returns
        -------
        np.ndarray
            Predicted samples.
        """
        fh_dates = self.fh_to_index(fh)
        fh_as_index = pd.Index(list(fh_dates.to_numpy()))

        if not isinstance(fh, ForecastingHorizon):
            fh = self._check_fh(fh)

        if X is None or X.shape[1] == 0:
            idx = reindex_time_series(self._y, fh_as_index).index
            X = pd.DataFrame(index=idx)
            X = self.aggregator_.transform(X)

        X_bottom = loc_bottom_series(X)

        if self._has_exogenous_variables:

            assert fh_as_index.isin(
                X_bottom.index.get_level_values(-1)
            ).all(), "Missing exogenous variables for some series or dates."
            if self.feature_transformer is not None:
                X_bottom = self.feature_transformer.transform(X_bottom)
            X_bottom = self.expand_columns_transformer_.transform(X_bottom)

        trend_data = self.trend_model_.transform(X=X_bottom, fh=fh_as_index)
        target_data = self.likelihood_model_.transform(X=None, fh=fh_as_index)

        exogenous_data = self._transform_effects(X=X_bottom, fh=fh_as_index)

        return dict(
            y=None,
            data=exogenous_data,
            trend_data=trend_data,
            target_data=target_data,
            **self.fit_and_predict_data_,
        )

    def predict_samples(
        self, fh: ForecastingHorizon, X: Optional[pd.DataFrame] = None
    ) -> np.ndarray:
        """Generate samples for the given exogenous variables and forecasting horizon.

        Parameters
        ----------
        X : pd.DataFrame, optional
            Exogenous variables.
        fh : ForecastingHorizon
            Forecasting horizon.

        Returns
        -------
        np.ndarray
            Predicted samples.
        """
        samples = super().predict_samples(X=X, fh=fh)

        return self.aggregator_.transform(samples)

    def _filter_series_tuples(self, levels: List[Tuple]) -> List[Tuple]:
        """Filter series tuples, returning only series of interest.

        Since this class performs a bottom-up aggregation, we are only interested in the
        bottom levels of the hierarchy. This method filters the series tuples, returning
        only the bottom levels.

        Parameters
        ----------
        levels : List[Tuple]
            The original levels of timeseries (`y.index.droplevel(-1).unique()`)

        Returns
        -------
        List[Tuple]
            The same object as `levels`, but with only the bottom levels.
        """
        # Make it a tuple for consistency
        if not isinstance(levels[0], (tuple, list)):
            levels = [(idx,) for idx in levels]

        bottom_levels = [idx for idx in levels if idx[-1] != "__total"]
        return bottom_levels

    @property
    def n_series(self):
        """Get the number of series.

        Returns
        -------
        int
            Number of series.
        """
        if self.internal_y_indexes_.nlevels == 1:
            return 1
        return len(
            self._filter_series_tuples(
                self.internal_y_indexes_.droplevel(-1).unique().tolist()
            )
        )

    def _postprocess_output(self, y: pd.DataFrame) -> pd.DataFrame:
        """Postprocess outputs, by aggregating them.

        Parameters
        ----------
        y : pd.DataFrame
            dataframe with output predictions

        Returns
        -------
        pd.DataFrame
            postprocessed dataframe

        """
        return self.aggregator_.transform(y)

    @classmethod
    def get_test_params(cls, parameter_set="default") -> List[dict[str, Any]]:
        """Params to be used in sktime unit tests.

        Parameters
        ----------
        parameter_set : str, optional
            The parameter set to be used (ignored in this implementation)

        Returns
        -------
        List[dict[str, int]]
            A list of dictionaries containing the test parameters.
        """
        from prophetverse.engine.prior import PriorPredictiveInferenceEngine
        from prophetverse.engine.mcmc import MCMCInferenceEngine
        from prophetverse.engine.optimizer import AdamOptimizer

        return [
            {"inference_engine": PriorPredictiveInferenceEngine(1)},
            {
                "inference_engine": MCMCInferenceEngine(
                    num_samples=1, num_warmup=1, num_chains=1
                ),
            },
        ]



"""LBFGS solver for NumPyro."""

from typing import Any, Callable, NamedTuple

import chex
import jax
import jax.numpy as jnp
import optax
import optax.tree_utils as otu
from numpyro.optim import _NumPyroOptim

from prophetverse.logger import logger


class LBFGS(_NumPyroOptim):
    """
    An L-BFGS optimizer integrated with NumPyro, using optax.

    Parameters
    ----------
    gtol : float, default=1e-6
        Gradient tolerance for stopping criterion.
    tol : float, default=1e-6
        Function value tolerance for stopping criterion.
    learning_rate : float, default=1e-3
        Initial learning rate.
    memory_size : int, default=10
        Memory size for L-BFGS updates.
    scale_init_precond : bool, default=True
        Whether to scale the initial preconditioner.
    max_linesearch_steps : int, default=20
        Maximum number of line search steps.
    initial_guess_strategy : str, default="one"
        Strategy for the initial line search step size guess.
    max_learning_rate : float, optional
        Maximum allowed learning rate during line search.
    linesearch_tol : float, default=0
        Tolerance parameter for line search.
    increase_factor : float, default=2
        Factor by which to increase step size during line search when conditions are
        met.
    slope_rtol : float, default=0.0001
        Relative tolerance for slope in the line search.
    curv_rtol : float, default=0.9
        Curvature condition tolerance for line search.
    approx_dec_rtol : float, default=0.000001
        Approximate decrease tolerance for line search.
    stepsize_precision : float, default=1e5
        Stepsize precision tolerance.
    """

    def __init__(
        self,
        max_iter: int = 100,
        gtol=1e-6,
        tol: float = 1e-6,
        learning_rate=1e-3,
        memory_size=10,
        scale_init_precond=True,
        # linesearch
        max_linesearch_steps=50,
        initial_guess_strategy="one",
        max_learning_rate=None,
        linesearch_tol=0,
        increase_factor=2,
        slope_rtol=0.0001,
        curv_rtol=0.9,
        approx_dec_rtol=0.000001,
        stepsize_precision=1e5,
    ):
        self.max_iter = max_iter
        self.gtol = gtol
        self.tol = tol
        self.learning_rate = learning_rate
        self.memory_size = memory_size
        self.scale_init_precond = scale_init_precond

        self.max_linesearch_steps = max_linesearch_steps
        self.initial_guess_strategy = initial_guess_strategy
        self.max_learning_rate = max_learning_rate
        self.linesearch_tol = linesearch_tol
        self.increase_factor = increase_factor
        self.slope_rtol = slope_rtol
        self.curv_rtol = curv_rtol
        self.approx_dec_rtol = approx_dec_rtol
        self.stepsize_precision = stepsize_precision

        self._transformation = self.get_transformation()

    def get_transformation(self) -> optax.GradientTransformationExtraArgs:
        """
        Construct the gradient transformation for L-BFGS optimization with line search.

        Returns
        -------
        optax.GradientTransformation
            An optax transformation that first prints iteration information,
            then applies L-BFGS updates with optional line search.
        """
        linesearch = optax.scale_by_zoom_linesearch(
            max_linesearch_steps=self.max_linesearch_steps,
            initial_guess_strategy="one",
            max_learning_rate=self.max_learning_rate,
            tol=self.linesearch_tol,
            increase_factor=self.increase_factor,
            slope_rtol=self.slope_rtol,
            curv_rtol=self.curv_rtol,
            approx_dec_rtol=self.approx_dec_rtol,
            stepsize_precision=self.stepsize_precision,
            verbose=False,
        )
        return optax.chain(
            print_info(),
            optax.lbfgs(
                learning_rate=self.learning_rate,
                memory_size=self.memory_size,
                scale_init_precond=self.scale_init_precond,
                linesearch=linesearch,
            ),
        )

    def eval_and_update(
        self,
        fn: Callable[[Any], tuple],
        state,
        forward_mode_differentiation=False,
    ):
        """
        Evaluate the function at the current parameters and perform the L-BFGS update.

        Parameters
        ----------
        fn : Callable[[Any], tuple]
            A function that given parameters returns a tuple (value, auxiliary_data).
        state : Any
            Current state of the optimization, including parameters and opt state.
        forward_mode_differentiation : bool, default=False
            Whether to use forward-mode differentiation (not currently used).

        Returns
        -------
        tuple
            A tuple ((value, grad), svi_state) where:
            - (value, grad) is the objective function value and gradients at the updated
                parameters.
            - svi_state is the updated SVI state, containing the updated parameters and
                optimizer state.
        """
        params = self.get_params(state)

        final_params, final_state, value, grad = run_opt(
            params,
            fun=fn,
            opt=self._transformation,
            max_iter=self.max_iter,
            gtol=self.gtol,
            tol=self.tol,
        )

        step_num = otu.tree_get(final_state, "count")
        svi_state = step_num, (final_params, final_state)

        return (value, None), svi_state

    def get_params(self, state):
        """Get the parameters from the state.

        This method is called from numpyro's SVI.
        """
        step, (params, opt_state) = state
        return params

    def init(self, params):
        """Initialize the optimizer state.

        This method is called from numpyro's SVI.
        """
        opt_state = self._transformation.init(params)
        return jnp.array(0), (params, opt_state)


def run_opt(init_params, fun, opt, max_iter, gtol, tol):
    """
    Run the optimization loop until convergence criteria are met or max iter reached.

    Parameters
    ----------
    init_params : Any
        Initial parameters for the optimization.
    fun : Callable[[Any], tuple]
        A function that given parameters returns a tuple (value, auxiliary_data).
    opt : optax.GradientTransformation
        The optax optimizer transformation.
    max_iter : int
        Maximum number of iterations to run the optimization.
    gtol : float
        Gradient norm tolerance for stopping criterion.
    tol : float
        Function value tolerance for stopping criterion.

    Returns
    -------
    final_params : Any
        The optimized parameters.
    final_state : Any
        The final state of the optimizer.
    final_value : chex.Numeric
        The final value of the objective function.
    grad : Any
        The gradient at the final parameters.
    """

    def _fun(params):
        value, _ = fun(params)
        return value

    value_and_grad_fun = optax.value_and_grad_from_state(_fun)

    def step(carry):
        params, state, _ = carry
        value, grad = value_and_grad_fun(params, state=state)
        updates, state = opt.update(
            grad, state, params, value=value, grad=grad, value_fn=_fun
        )
        params = optax.apply_updates(params, updates)
        return params, state, value

    def continuing_criterion(carry):
        _, state, value = carry
        iter_num = otu.tree_get(state, "count")
        grad = otu.tree_get(state, "grad")
        err = otu.tree_l2_norm(grad)
        return (iter_num == 0) | (
            (iter_num < max_iter) & (err >= gtol) & (value >= tol)
        )

    init_carry = (init_params, opt.init(init_params), jnp.inf)
    final_params, final_state, final_value = jax.lax.while_loop(
        continuing_criterion, step, init_carry
    )

    _, grad = value_and_grad_fun(final_params, state=final_state)
    return final_params, final_state, final_value, grad


class InfoState(NamedTuple):
    """
    A named tuple holding iteration information.

    Attributes
    ----------
    iter_num : chex.Numeric
        The current iteration number.
    """

    iter_num: chex.Numeric


def print_info():
    """
    Create a gradient transformation.

    Prints iteration number, function value,
    and gradient norm at each step. This is useful for debugging and monitoring
    the optimization process.

    Returns
    -------
    optax.GradientTransformationExtraArgs
        A gradient transformation that, when applied, prints optimization info
        at each iteration.
    """

    def init_fn(params):
        del params
        return InfoState(iter_num=0)

    def update_fn(updates, state, params, *, value, grad, **extra_args):

        del params, extra_args

        fmt = "Iteration: {i}, Value: {v}, Gradient norm: {e}"

        def _log(*args, **kwargs):
            logger.debug(fmt.format(*args, **kwargs))

        jax.debug.callback(_log, i=state.iter_num, v=value, e=otu.tree_l2_norm(grad))

        return updates, InfoState(iter_num=state.iter_num + 1)

    return optax.GradientTransformationExtraArgs(init_fn, update_fn)
